"""
Manage RDP Service on Windows servers
"""

import logging
import re

import salt.utils.platform
from salt.utils.decorators import depends

try:
    import win32ts
    from pywintypes import error as PyWinError

    _HAS_WIN32TS_DEPENDENCIES = True
except ImportError:
    _HAS_WIN32TS_DEPENDENCIES = False

_LOG = logging.getLogger(__name__)


def __virtual__():
    """
    Only works on Windows systems
    """
    if salt.utils.platform.is_windows():
        return "rdp"
    return (False, "Module only works on Windows.")


def _parse_return_code_powershell(string):
    """
    return from the input string the return code of the powershell command
    """

    regex = re.search(r"ReturnValue\s*: (\d*)", string)
    if not regex:
        return (False, "Could not parse PowerShell return code.")
    else:
        return int(regex.group(1))


def _psrdp(cmd):
    """
    Create a Win32_TerminalServiceSetting WMI Object as $RDP and execute the
    command cmd returns the STDOUT of the command
    """
    rdp = (
        "$RDP = Get-WmiObject -Class Win32_TerminalServiceSetting "
        "-Namespace root\\CIMV2\\TerminalServices -Computer . "
        "-Authentication 6 -ErrorAction Stop"
    )
    return __salt__["cmd.run"](
        "{} ; {}".format(rdp, cmd), shell="powershell", python_shell=True
    )


def enable():
    """
    Enable RDP the service on the server

    CLI Example:

    .. code-block:: bash

        salt '*' rdp.enable
    """

    return _parse_return_code_powershell(_psrdp("$RDP.SetAllowTsConnections(1,1)")) == 0


def disable():
    """
    Disable RDP the service on the server

    CLI Example:

    .. code-block:: bash

        salt '*' rdp.disable
    """

    return _parse_return_code_powershell(_psrdp("$RDP.SetAllowTsConnections(0,1)")) == 0


def status():
    """
    Show if rdp is enabled on the server

    CLI Example:

    .. code-block:: bash

        salt '*' rdp.status
    """

    out = int(_psrdp("echo $RDP.AllowTSConnections").strip())
    return out != 0


@depends(_HAS_WIN32TS_DEPENDENCIES)
def list_sessions(logged_in_users_only=False):
    """
    List information about the sessions.

    .. versionadded:: 2016.11.0

    :param logged_in_users_only: If True, only return sessions with users logged in.
    :return: A list containing dictionaries of session information.

    CLI Example:

    .. code-block:: bash

        salt '*' rdp.list_sessions
    """
    ret = list()
    server = win32ts.WTS_CURRENT_SERVER_HANDLE
    protocols = {
        win32ts.WTS_PROTOCOL_TYPE_CONSOLE: "console",
        win32ts.WTS_PROTOCOL_TYPE_ICA: "citrix",
        win32ts.WTS_PROTOCOL_TYPE_RDP: "rdp",
    }
    statuses = {
        win32ts.WTSActive: "active",
        win32ts.WTSConnected: "connected",
        win32ts.WTSConnectQuery: "connect_query",
        win32ts.WTSShadow: "shadow",
        win32ts.WTSDisconnected: "disconnected",
        win32ts.WTSIdle: "idle",
        win32ts.WTSListen: "listen",
        win32ts.WTSReset: "reset",
        win32ts.WTSDown: "down",
        win32ts.WTSInit: "init",
    }

    for session in win32ts.WTSEnumerateSessions(server):
        user = (
            win32ts.WTSQuerySessionInformation(
                server, session["SessionId"], win32ts.WTSUserName
            )
            or None
        )
        protocol_id = win32ts.WTSQuerySessionInformation(
            server, session["SessionId"], win32ts.WTSClientProtocolType
        )
        status_id = win32ts.WTSQuerySessionInformation(
            server, session["SessionId"], win32ts.WTSConnectState
        )
        protocol = protocols.get(protocol_id, "unknown")
        connection_status = statuses.get(status_id, "unknown")
        station = session["WinStationName"] or "Disconnected"
        connection_info = {
            "connection_status": connection_status,
            "protocol": protocol,
            "session_id": session["SessionId"],
            "station": station,
            "user": user,
        }
        if logged_in_users_only:
            if user:
                ret.append(connection_info)
        else:
            ret.append(connection_info)

    if not ret:
        _LOG.warning("No sessions found.")
    return sorted(ret, key=lambda k: k["session_id"])


@depends(_HAS_WIN32TS_DEPENDENCIES)
def get_session(session_id):
    """
    Get information about a session.

    .. versionadded:: 2016.11.0

    :param session_id: The numeric Id of the session.
    :return: A dictionary of session information.

    CLI Example:

    .. code-block:: bash

        salt '*' rdp.get_session session_id

        salt '*' rdp.get_session 99
    """
    ret = dict()
    sessions = list_sessions()
    session = [item for item in sessions if item["session_id"] == session_id]

    if session:
        ret = session[0]

    if not ret:
        _LOG.warning("No session found for id: %s", session_id)
    return ret


@depends(_HAS_WIN32TS_DEPENDENCIES)
def disconnect_session(session_id):
    """
    Disconnect a session.

    .. versionadded:: 2016.11.0

    :param session_id: The numeric Id of the session.
    :return: A boolean representing whether the disconnect succeeded.

    CLI Example:

    .. code-block:: bash

        salt '*' rdp.disconnect_session session_id

        salt '*' rdp.disconnect_session 99
    """
    try:
        win32ts.WTSDisconnectSession(
            win32ts.WTS_CURRENT_SERVER_HANDLE, session_id, True
        )
    except PyWinError as error:
        _LOG.error("Error calling WTSDisconnectSession: %s", error)
        return False
    return True


@depends(_HAS_WIN32TS_DEPENDENCIES)
def logoff_session(session_id):
    """
    Initiate the logoff of a session.

    .. versionadded:: 2016.11.0

    :param session_id: The numeric Id of the session.
    :return: A boolean representing whether the logoff succeeded.

    CLI Example:

    .. code-block:: bash

        salt '*' rdp.logoff_session session_id

        salt '*' rdp.logoff_session 99
    """
    try:
        win32ts.WTSLogoffSession(win32ts.WTS_CURRENT_SERVER_HANDLE, session_id, True)
    except PyWinError as error:
        _LOG.error("Error calling WTSLogoffSession: %s", error)
        return False
    return True



"""
Tests for the salt runner

.. versionadded:: 2016.11.0
"""

import pytest

from tests.support.case import ShellCase


@pytest.mark.windows_whitelisted
class SaltRunnerTest(ShellCase):
    """
    Test the salt runner
    """

    @pytest.mark.slow_test
    def test_salt_cmd(self):
        """
        test return values of salt.cmd
        """
        ret = self.run_run_plus("salt.cmd", "test.ping")
        out_ret = ret.get("out")[0]
        return_ret = ret.get("return")

        self.assertEqual(out_ret, "True")
        self.assertTrue(return_ret)

    @pytest.mark.slow_test
    def test_salt_cmd_invalid(self):
        """
        test return values of salt.cmd invalid parameters
        """
        ret = self.run_run_plus("salt.cmd")
        expected = "Passed invalid arguments:"
        self.assertIn(expected, ret["return"])



#!/usr/bin/env python
# pylint: skip-file

from __future__ import absolute_import, division, print_function

from salt.ext.tornado import gen, ioloop
from salt.ext.tornado.log import app_log
from salt.ext.tornado.testing import AsyncTestCase, gen_test, ExpectLog
from salt.ext.tornado.test.util import unittest, skipBefore35, exec_test
import contextlib
import os
import traceback
import warnings


@contextlib.contextmanager
def set_environ(name, value):
    old_value = os.environ.get(name)
    os.environ[name] = value

    try:
        yield
    finally:
        if old_value is None:
            del os.environ[name]
        else:
            os.environ[name] = old_value


class AsyncTestCaseTest(AsyncTestCase):
    def test_exception_in_callback(self):
        self.io_loop.add_callback(lambda: 1 / 0)
        try:
            self.wait()
            self.fail("did not get expected exception")
        except ZeroDivisionError:
            pass

    def test_wait_timeout(self):
        time = self.io_loop.time

        # Accept default 5-second timeout, no error
        self.io_loop.add_timeout(time() + 0.01, self.stop)
        self.wait()

        # Timeout passed to wait()
        self.io_loop.add_timeout(time() + 1, self.stop)
        with self.assertRaises(self.failureException):
            self.wait(timeout=0.01)

        # Timeout set with environment variable
        self.io_loop.add_timeout(time() + 1, self.stop)
        with set_environ('ASYNC_TEST_TIMEOUT', '0.01'):
            with self.assertRaises(self.failureException):
                self.wait()

    def test_subsequent_wait_calls(self):
        """
        This test makes sure that a second call to wait()
        clears the first timeout.
        """
        self.io_loop.add_timeout(self.io_loop.time() + 0.00, self.stop)
        self.wait(timeout=0.02)
        self.io_loop.add_timeout(self.io_loop.time() + 0.03, self.stop)
        self.wait(timeout=0.15)

    def test_multiple_errors(self):
        def fail(message):
            raise Exception(message)
        self.io_loop.add_callback(lambda: fail("error one"))
        self.io_loop.add_callback(lambda: fail("error two"))
        # The first error gets raised; the second gets logged.
        with ExpectLog(app_log, "multiple unhandled exceptions"):
            with self.assertRaises(Exception) as cm:
                self.wait()
        self.assertEqual(str(cm.exception), "error one")


class AsyncTestCaseWrapperTest(unittest.TestCase):
    def test_undecorated_generator(self):
        class Test(AsyncTestCase):
            def test_gen(self):
                yield
        test = Test('test_gen')
        result = unittest.TestResult()
        test.run(result)
        self.assertEqual(len(result.errors), 1)
        self.assertIn("should be decorated", result.errors[0][1])

    @skipBefore35
    def test_undecorated_coroutine(self):
        namespace = exec_test(globals(), locals(), """
        class Test(AsyncTestCase):
            async def test_coro(self):
                pass
        """)

        test_class = namespace['Test']
        test = test_class('test_coro')
        result = unittest.TestResult()

        # Silence "RuntimeWarning: coroutine 'test_coro' was never awaited".
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            test.run(result)

        self.assertEqual(len(result.errors), 1)
        self.assertIn("should be decorated", result.errors[0][1])

    def test_undecorated_generator_with_skip(self):
        class Test(AsyncTestCase):
            @unittest.skip("don't run this")
            def test_gen(self):
                yield
        test = Test('test_gen')
        result = unittest.TestResult()
        test.run(result)
        self.assertEqual(len(result.errors), 0)
        self.assertEqual(len(result.skipped), 1)

    def test_other_return(self):
        class Test(AsyncTestCase):
            def test_other_return(self):
                return 42
        test = Test('test_other_return')
        result = unittest.TestResult()
        test.run(result)
        self.assertEqual(len(result.errors), 1)
        self.assertIn("Return value from test method ignored", result.errors[0][1])


class SetUpTearDownTest(unittest.TestCase):
    def test_set_up_tear_down(self):
        """
        This test makes sure that AsyncTestCase calls super methods for
        setUp and tearDown.

        InheritBoth is a subclass of both AsyncTestCase and
        SetUpTearDown, with the ordering so that the super of
        AsyncTestCase will be SetUpTearDown.
        """
        events = []
        result = unittest.TestResult()

        class SetUpTearDown(unittest.TestCase):
            def setUp(self):
                events.append('setUp')

            def tearDown(self):
                events.append('tearDown')

        class InheritBoth(AsyncTestCase, SetUpTearDown):
            def test(self):
                events.append('test')

        InheritBoth('test').run(result)
        expected = ['setUp', 'test', 'tearDown']
        self.assertEqual(expected, events)


class GenTest(AsyncTestCase):
    def setUp(self):
        super(GenTest, self).setUp()
        self.finished = False

    def tearDown(self):
        self.assertTrue(self.finished)
        super(GenTest, self).tearDown()

    @gen_test
    def test_sync(self):
        self.finished = True

    @gen_test
    def test_async(self):
        yield gen.Task(self.io_loop.add_callback)
        self.finished = True

    def test_timeout(self):
        # Set a short timeout and exceed it.
        @gen_test(timeout=0.1)
        def test(self):
            yield gen.Task(self.io_loop.add_timeout, self.io_loop.time() + 1)

        # This can't use assertRaises because we need to inspect the
        # exc_info triple (and not just the exception object)
        try:
            test(self)
            self.fail("did not get expected exception")
        except ioloop.TimeoutError:
            # The stack trace should blame the add_timeout line, not just
            # unrelated IOLoop/testing internals.
            self.assertIn(
                "gen.Task(self.io_loop.add_timeout, self.io_loop.time() + 1)",
                traceback.format_exc())

        self.finished = True

    def test_no_timeout(self):
        # A test that does not exceed its timeout should succeed.
        @gen_test(timeout=1)
        def test(self):
            time = self.io_loop.time
            yield gen.Task(self.io_loop.add_timeout, time() + 0.1)

        test(self)
        self.finished = True

    def test_timeout_environment_variable(self):
        @gen_test(timeout=0.5)
        def test_long_timeout(self):
            time = self.io_loop.time
            yield gen.Task(self.io_loop.add_timeout, time() + 0.25)

        # Uses provided timeout of 0.5 seconds, doesn't time out.
        with set_environ('ASYNC_TEST_TIMEOUT', '0.1'):
            test_long_timeout(self)

        self.finished = True

    def test_no_timeout_environment_variable(self):
        @gen_test(timeout=0.01)
        def test_short_timeout(self):
            time = self.io_loop.time
            yield gen.Task(self.io_loop.add_timeout, time() + 1)

        # Uses environment-variable timeout of 0.1, times out.
        with set_environ('ASYNC_TEST_TIMEOUT', '0.1'):
            with self.assertRaises(ioloop.TimeoutError):
                test_short_timeout(self)

        self.finished = True

    def test_with_method_args(self):
        @gen_test
        def test_with_args(self, *args):
            self.assertEqual(args, ('test',))
            yield gen.Task(self.io_loop.add_callback)

        test_with_args(self, 'test')
        self.finished = True

    def test_with_method_kwargs(self):
        @gen_test
        def test_with_kwargs(self, **kwargs):
            self.assertDictEqual(kwargs, {'test': 'test'})
            yield gen.Task(self.io_loop.add_callback)

        test_with_kwargs(self, test='test')
        self.finished = True

    @skipBefore35
    def test_native_coroutine(self):
        namespace = exec_test(globals(), locals(), """
        @gen_test
        async def test(self):
            self.finished = True
        """)

        namespace['test'](self)

    @skipBefore35
    def test_native_coroutine_timeout(self):
        # Set a short timeout and exceed it.
        namespace = exec_test(globals(), locals(), """
        @gen_test(timeout=0.1)
        async def test(self):
            await gen.sleep(1)
        """)

        try:
            namespace['test'](self)
            self.fail("did not get expected exception")
        except ioloop.TimeoutError:
            self.finished = True


if __name__ == '__main__':
    unittest.main()



import os.path
import shutil
import subprocess

import psutil
import pytest

import salt.exceptions

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def dsc(modules):
    # We seem to be hitting an issue where there is a consistency check in
    # progress during some of the tests. When this happens, the test fails
    # This should disabled background refreshes
    # https://github.com/saltstack/salt/issues/62714
    existing_config_mode = modules.dsc.get_lcm_config()["ConfigurationMode"]
    modules.dsc.set_lcm_config(config_mode="ApplyOnly")
    yield modules.dsc
    modules.dsc.set_lcm_config(config_mode=existing_config_mode)


@pytest.fixture(scope="function")
def ps1_file():
    """
    This will create a DSC file to be configured. When configured it will create
    a localhost.mof file in the `HelloWorld` directory in Temp
    """
    ps1_contents = r"""
    Configuration HelloWorld {

        # Import the module that contains the File resource.
        Import-DscResource -ModuleName PsDesiredStateConfiguration

        # The Node statement specifies which targets to compile MOF files for, when this configuration is executed.
        Node ("localhost") {

            # The File resource can ensure the state of files, or copy them from a source to a destination with persistent updates.
            File HelloWorld {
                DestinationPath = "C:\Temp\HelloWorld.txt"
                Ensure          = "Present"
                Contents        = "Hello World, ps1_file"
            }
        }
    }
    """
    with pytest.helpers.temp_file("hello_world.ps1", contents=ps1_contents) as file:
        yield file
    if os.path.exists(file.parent / "HelloWorld"):
        shutil.rmtree(file.parent / "HelloWorld")
    if os.path.exists(file):
        os.remove(file)


@pytest.fixture(scope="function")
def ps1_file_multiple():
    """
    This will create a DSC file to be configured. When configured it will create
    a localhost.mof file in the `HelloWorld2` directory in Temp
    """
    ps1_contents = r"""
    Configuration HelloWorldMultiple {

        # Import the module that contains the File resource.
        Import-DscResource -ModuleName PsDesiredStateConfiguration

        # The Node statement specifies which targets to compile MOF files for, when this configuration is executed.
        Node ("localhost") {

            # The File resource can ensure the state of files, or copy them from a source to a destination with persistent updates.
            File HelloWorld {
                DestinationPath = "C:\Temp\HelloWorld.txt"
                Ensure          = "Present"
                Contents        = "Hello World from DSC!"
            }

            # The File resource can ensure the state of files, or copy them from a source to a destination with persistent updates.
            File HelloWorld2 {
                DestinationPath = "C:\Temp\HelloWorld2.txt"
                Ensure          = "Present"
                Contents        = "Hello World, ps1_file_multiple"
            }
        }
    }
    """

    with pytest.helpers.temp_file(
        "hello_world_multiple.ps1", contents=ps1_contents
    ) as file:
        yield file
    if os.path.exists(file.parent / "HelloWorldMultiple"):
        shutil.rmtree(file.parent / "HelloWorldMultiple")
    if os.path.exists(file):
        os.remove(file)


@pytest.fixture(scope="function")
def ps1_file_meta():
    """
    This will create a DSC file to be configured. When configured it will create
    a localhost.mof file and a localhost.meta.mof file in the `HelloWorld`
    directory in Temp
    """
    ps1_contents = r"""
    Configuration HelloWorld {

        # Import the module that contains the File resource.
        Import-DscResource -ModuleName PsDesiredStateConfiguration

        # The Node statement specifies which targets to compile MOF files for, when this configuration is executed.
        Node ("localhost") {

            # The File resource can ensure the state of files, or copy them from a source to a destination with persistent updates.
            File HelloWorld {
                DestinationPath = "C:\Temp\HelloWorld.txt"
                Ensure          = "Present"
                Contents        = "Hello World, ps1_file_meta "
            }

            # Set some Meta Config
            LocalConfigurationManager {
                ConfigurationMode  = "ApplyAndMonitor"
                RebootNodeIfNeeded = $false
                RefreshMode        = "PUSH"
            }
        }
    }
    """
    with pytest.helpers.temp_file("test.ps1", contents=ps1_contents) as file:
        yield file
    if os.path.exists(file.parent / "HelloWorld"):
        shutil.rmtree(file.parent / "HelloWorld")
    if os.path.exists(file):
        os.remove(file)


@pytest.fixture(scope="module")
def psd1_file():
    """
    This will create a config data file to be applied with the config file in
    Temp
    """
    psd1_contents = r"""
    @{
        AllNodes = @(
            @{
                NodeName = 'localhost'
                PSDscAllowPlainTextPassword = $true
                PSDscAllowDomainUser = $true
            }
        )
    }
    """
    with pytest.helpers.temp_file("test.psd1", contents=psd1_contents) as file:
        yield file
    if os.path.exists(file):
        os.remove(file)


def test_compile_config_missing(dsc):
    path = "C:\\Path\\not\\exists.ps1"
    with pytest.raises(salt.exceptions.CommandExecutionError) as exc:
        dsc.compile_config(path=path)
    assert exc.value.message == "{} not found".format(path)


@pytest.mark.destructive_test
def test_compile_config(dsc, ps1_file, psd1_file):
    """
    Test compiling a simple config
    """
    dsc.remove_config(reset=False)
    result = dsc.compile_config(
        path=str(ps1_file),
        config_name="HelloWorld",
        config_data=str(psd1_file),
    )
    assert isinstance(result, dict)
    assert result["Exists"] is True


@pytest.mark.destructive_test
def test_compile_config_issue_61261(dsc, ps1_file_meta, psd1_file):
    """
    Test compiling a config that includes meta data
    """
    dsc.remove_config(reset=False)
    result = dsc.compile_config(
        path=str(ps1_file_meta),
        config_name="HelloWorld",
        config_data=str(psd1_file),
    )
    assert isinstance(result, dict)
    assert result["Exists"] is True


def test_apply_config_missing(dsc):
    path = "C:\\Path\\not\\exists"
    with pytest.raises(salt.exceptions.CommandExecutionError) as exc:
        dsc.apply_config(path=path)
    assert exc.value.message == "{} not found".format(path)


@pytest.mark.destructive_test
def test_apply_config(dsc, ps1_file, psd1_file):
    """
    Test applying a simple config
    """
    dsc.remove_config(reset=False)
    dsc.compile_config(
        path=str(ps1_file),
        config_name="HelloWorld",
        config_data=str(psd1_file),
    )
    result = dsc.apply_config(path=ps1_file.parent / "HelloWorld")
    assert result is True


def test_get_config_not_configured(dsc):
    dsc.remove_config(reset=False)
    with pytest.raises(salt.exceptions.CommandExecutionError) as exc:
        dsc.get_config()
    assert exc.value.message == "Not Configured"


def test_get_config_single(dsc, ps1_file, psd1_file):
    dsc.remove_config(reset=False)
    dsc.run_config(
        path=str(ps1_file),
        config_name="HelloWorld",
        config_data=str(psd1_file),
    )
    result = dsc.get_config()
    assert "HelloWorld" in result
    assert "[File]HelloWorld" in result["HelloWorld"]
    assert "DestinationPath" in result["HelloWorld"]["[File]HelloWorld"]


def test_get_config_multiple(dsc, ps1_file_multiple, psd1_file):
    dsc.remove_config(reset=False)
    dsc.run_config(
        path=str(ps1_file_multiple),
        config_name="HelloWorldMultiple",
        config_data=str(psd1_file),
    )
    result = dsc.get_config()
    assert "HelloWorldMultiple" in result
    assert "[File]HelloWorld" in result["HelloWorldMultiple"]
    assert "DestinationPath" in result["HelloWorldMultiple"]["[File]HelloWorld"]
    assert "[File]HelloWorld2" in result["HelloWorldMultiple"]
    assert "DestinationPath" in result["HelloWorldMultiple"]["[File]HelloWorld2"]


def _reset_config(dsc):
    """
    Resets the DSC config. If files are locked, this will attempt to kill the
    all running WmiPrvSE processes. Windows will respawn the ones it needs
    """
    tries = 1
    while True:
        try:
            tries += 1
            dsc.remove_config(reset=True)
            break
        except salt.exceptions.CommandExecutionError:
            if tries > 12:
                raise

            # Kill the processes
            proc_name = "wmiprvse.exe"
            for proc in psutil.process_iter():
                if proc.name().lower() == proc_name:
                    proc.kill()

            continue


def test_get_config_status_not_configured(dsc):
    _reset_config(dsc)
    with pytest.raises(salt.exceptions.CommandExecutionError) as exc:
        dsc.get_config_status()
    assert exc.value.message == "Not Configured"


def test_get_config_status(dsc, ps1_file, psd1_file):
    dsc.remove_config(reset=False)
    dsc.run_config(
        path=str(ps1_file),
        config_name="HelloWorld",
        config_data=str(psd1_file),
    )
    result = dsc.get_config_status()
    assert "MetaData" in result
    assert "HelloWorld" in result["MetaData"]
    assert result["Status"] == "Success"


def test_test_config_not_configured(dsc):
    subprocess.run(
        ["cmd", "/c", "winrm", "quickconfig", "-quiet"],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        check=True,
    )
    dsc.remove_config(reset=False)
    with pytest.raises(salt.exceptions.CommandExecutionError) as exc:
        dsc.test_config()
    assert exc.value.message == "Not Configured"


def test_test_config(dsc, ps1_file, psd1_file):
    dsc.remove_config(reset=False)
    dsc.run_config(
        path=str(ps1_file),
        config_name="HelloWorld",
        config_data=str(psd1_file),
    )
    result = dsc.test_config()
    assert result is True


def test_get_lcm_config(dsc):
    config_items = [
        "ConfigurationModeFrequencyMins",
        "LCMState",
        "RebootNodeIfNeeded",
        "ConfigurationMode",
        "ActionAfterReboot",
        "RefreshMode",
        "CertificateID",
        "ConfigurationID",
        "RefreshFrequencyMins",
        "AllowModuleOverwrite",
        "DebugMode",
        "StatusRetentionTimeInDays",
    ]
    dsc.remove_config(reset=False)
    result = dsc.get_lcm_config()
    for item in config_items:
        assert item in result


def test_set_lcm_config(dsc):
    current = dsc.get_lcm_config()["ConfigurationMode"]
    dsc.set_lcm_config(config_mode="ApplyOnly")
    try:
        results = dsc.get_lcm_config()
        assert results["ConfigurationMode"] == "ApplyOnly"
    finally:
        dsc.set_lcm_config(config_mode=current)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.moosefs as moosefs
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {moosefs: {}}


def test_dirinfo():
    """
    Test if it return information on a directory located on the Moose
    """
    mock = MagicMock(return_value={"stdout": "Salt:salt"})
    with patch.dict(moosefs.__salt__, {"cmd.run_all": mock}):
        assert moosefs.dirinfo("/tmp/salt") == {"Salt": "salt"}


def test_fileinfo():
    """
    Test if it returns information on a file located on the Moose
    """
    mock = MagicMock(return_value={"stdout": ""})
    with patch.dict(moosefs.__salt__, {"cmd.run_all": mock}):
        assert moosefs.fileinfo("/tmp/salt") == {}


def test_mounts():
    """
    Test if it returns a list of current MooseFS mounts
    """
    mock = MagicMock(return_value={"stdout": ""})
    with patch.dict(moosefs.__salt__, {"cmd.run_all": mock}):
        assert moosefs.mounts() == {}


def test_getgoal():
    """
    Test if it returns goal(s) for a file or directory
    """
    mock = MagicMock(return_value={"stdout": "Salt: salt"})
    with patch.dict(moosefs.__salt__, {"cmd.run_all": mock}):
        assert moosefs.getgoal("/tmp/salt") == {"goal": "salt"}



import pytest

import salt.modules.selinux as selinux
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {selinux: {}}


def test_fcontext_get_policy_parsing():
    """
    Test to verify that the parsing of the semanage output into fields is
    correct. Added with #45784.
    """
    cases = [
        {
            "semanage_out": (
                "/var/www(/.*)?     all files         "
                " system_u:object_r:httpd_sys_content_t:s0 "
            ),
            "name": "/var/www(/.*)?",
            "filetype": "all files",
            "sel_user": "system_u",
            "sel_role": "object_r",
            "sel_type": "httpd_sys_content_t",
            "sel_level": "s0",
        },
        {
            "semanage_out": (
                "/var/www(/.*)? all files         "
                " system_u:object_r:httpd_sys_content_t:s0  "
            ),
            "name": "/var/www(/.*)?",
            "filetype": "all files",
            "sel_user": "system_u",
            "sel_role": "object_r",
            "sel_type": "httpd_sys_content_t",
            "sel_level": "s0",
        },
        {
            "semanage_out": (
                "/var/lib/dhcp3?                                    directory      "
                "    system_u:object_r:dhcp_state_t:s0	"
            ),
            "name": "/var/lib/dhcp3?",
            "filetype": "directory",
            "sel_user": "system_u",
            "sel_role": "object_r",
            "sel_type": "dhcp_state_t",
            "sel_level": "s0",
        },
        {
            "semanage_out": (
                "/var/lib/dhcp3?  directory         "
                " system_u:object_r:dhcp_state_t:s0"
            ),
            "name": "/var/lib/dhcp3?",
            "filetype": "directory",
            "sel_user": "system_u",
            "sel_role": "object_r",
            "sel_type": "dhcp_state_t",
            "sel_level": "s0",
        },
        {
            "semanage_out": (
                "/var/lib/dhcp3? directory         "
                " system_u:object_r:dhcp_state_t:s0"
            ),
            "name": "/var/lib/dhcp3?",
            "filetype": "directory",
            "sel_user": "system_u",
            "sel_role": "object_r",
            "sel_type": "dhcp_state_t",
            "sel_level": "s0",
        },
    ]

    for case in cases:
        with patch.dict(
            selinux.__salt__,
            {"cmd.shell": MagicMock(return_value=case["semanage_out"])},
        ):
            ret = selinux.fcontext_get_policy(case["name"])
            assert ret["filespec"] == case["name"]
            assert ret["filetype"] == case["filetype"]
            assert ret["sel_user"] == case["sel_user"]
            assert ret["sel_role"] == case["sel_role"]
            assert ret["sel_type"] == case["sel_type"]
            assert ret["sel_level"] == case["sel_level"]


@pytest.mark.parametrize(
    "name, protocol, port, expected",
    (
        ("tcp/80", None, None, ("tcp", "80")),
        ("udp/53", None, None, ("udp", "53")),
        ("tcp_test_dns", "tcp", "53", ("tcp", "53")),
        ("udp_test/dns", "udp", "53", ("udp", "53")),
    ),
)
def test_parse_protocol_port_positive(name, protocol, port, expected):
    """
    Test to verify positive parsing name, protocol and port combinations
    """
    ret = selinux._parse_protocol_port(name, protocol, port)
    assert ret == expected


@pytest.mark.parametrize(
    "name, protocol, port",
    (
        ("invalid_name_no_args", None, None),
        ("invalid_proto/80", "nottcp", "80"),
        ("invalid_port", "tcp", "notaport"),
        ("missing_proto", None, "80"),
        ("missing_port", "udp", None),
    ),
)
def test_parse_protocol_port_negative(name, protocol, port):
    """
    Test to verify negative parsing of name, protocol and port combinations
    """
    pytest.raises(
        SaltInvocationError,
        selinux._parse_protocol_port,
        name,
        protocol,
        port,
    )


def test_port_get_policy_parsing():
    """
    Test to verify that the parsing of the semanage port output into fields is correct.
    """
    cases = [
        {
            "semanage_out": "cma_port_t                     tcp      1050",
            "name": "tcp/1050",
            "expected": {
                "sel_type": "cma_port_t",
                "protocol": "tcp",
                "port": "1050",
            },
        },
        {
            "semanage_out": (
                "cluster_port_t                 tcp      5149, 40040, 50006-50008"
            ),
            "name": "tcp/40040",
            "expected": {
                "sel_type": "cluster_port_t",
                "protocol": "tcp",
                "port": "5149, 40040, 50006-50008",
            },
        },
        {
            "semanage_out": (
                "http_port_t                    tcp      9008, 8010, 9002-9003, 80,"
                " 81, 443, 488, 8008, 8009, 8443, 9000"
            ),
            "name": "tcp/9000",
            "expected": {
                "sel_type": "http_port_t",
                "protocol": "tcp",
                "port": (
                    "9008, 8010, 9002-9003, 80, 81, 443, 488, 8008, 8009, 8443," " 9000"
                ),
            },
        },
        {
            "semanage_out": (
                "vnc_port_t                     tcp      5985-5999, 5900-5983"
            ),
            "name": "tcp/5985-5999",
            "expected": {
                "sel_type": "vnc_port_t",
                "protocol": "tcp",
                "port": "5985-5999, 5900-5983",
            },
        },
        {
            "semanage_out": (
                "zebra_port_t                   tcp      2606, 2608-2609, 2600-2604"
            ),
            "name": "tcp/2608-2609",
            "expected": {
                "sel_type": "zebra_port_t",
                "protocol": "tcp",
                "port": "2606, 2608-2609, 2600-2604",
            },
        },
        {
            "semanage_out": (
                "radius_port_t                  udp      1645, 1812, 18120-18121"
            ),
            "name": "tcp/18120-18121",
            "expected": {
                "sel_type": "radius_port_t",
                "protocol": "udp",
                "port": "1645, 1812, 18120-18121",
            },
        },
    ]

    for case in cases:
        with patch.dict(
            selinux.__salt__,
            {"cmd.shell": MagicMock(return_value=case["semanage_out"])},
        ):
            ret = selinux.port_get_policy(case["name"])
            assert ret == case["expected"]


def test_fcontext_policy_parsing_new():
    """
    Test parsing the stdout response of restorecon used in fcontext_policy_applied, new style.
    """
    restorecon_ret = (
        "Would relabel /foo/bar from some_u:some_r:some_t:s0 to"
        " other_u:other_r:other_t:s0"
    )
    with patch.object(
        selinux, "fcontext_policy_is_applied", return_value=restorecon_ret
    ), patch.dict(
        selinux.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
    ):
        assert selinux.fcontext_apply_policy("/foo/bar") == {
            "changes": {
                "/foo/bar": {
                    "old": {
                        "sel_role": "some_r",
                        "sel_type": "some_t",
                        "sel_user": "some_u",
                    },
                    "new": {
                        "sel_role": "other_r",
                        "sel_type": "other_t",
                        "sel_user": "other_u",
                    },
                },
            },
            "retcode": 0,
        }


def test_fcontext_policy_parsing_old():
    """
    Test parsing the stdout response of restorecon used in fcontext_policy_applied, old style.
    """
    restorecon_ret = (
        "restorecon reset /foo/bar context"
        " some_u:some_r:some_t:s0->other_u:other_r:other_t:s0"
    )
    with patch.object(
        selinux, "fcontext_policy_is_applied", return_value=restorecon_ret
    ), patch.dict(
        selinux.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
    ):
        assert selinux.fcontext_apply_policy("/foo/bar") == {
            "changes": {
                "/foo/bar": {
                    "old": {
                        "sel_role": "some_r",
                        "sel_type": "some_t",
                        "sel_user": "some_u",
                    },
                    "new": {
                        "sel_role": "other_r",
                        "sel_type": "other_t",
                        "sel_user": "other_u",
                    },
                },
            },
            "retcode": 0,
        }


def test_fcontext_policy_parsing_fail():
    """
    Test failure response for invalid restorecon data.
    """
    restorecon_ret = "And now for something completely different."
    with patch.object(
        selinux, "fcontext_policy_is_applied", return_value=restorecon_ret
    ), patch.dict(
        selinux.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
    ):
        assert selinux.fcontext_apply_policy("/foo/bar") == {
            "retcode": 1,
            "error": "Unrecognized response from restorecon command.",
        }


def test_selinux_config_enforcing():
    """
    Test values written to /etc/selinux/config are lowercase
    """
    mock_file = """
# This file controls the state of SELinux on the system.
# SELINUX= can take one of these three values:
#     enforcing - SELinux security policy is enforced.
#     permissive - SELinux prints warnings instead of enforcing.
#     disabled - No SELinux policy is loaded.
## SELINUX=disabled
SELINUX=permissive
# SELINUXTYPE= can take one of these three values:
#     targeted - Targeted processes are protected,
#     minimum - Modification of targeted policy. Only selected processes are protected.
#     mls - Multi Level Security protection.
SELINUXTYPE=targeted

"""
    with patch("salt.utils.files.fopen", mock_open(read_data=mock_file)) as m_open:
        selinux.setenforce("Enforcing")
        writes = m_open.write_calls()
        assert writes
        for line in writes:
            if line.startswith("SELINUX="):
                assert line == "SELINUX=enforcing"


def test_selinux_config_permissive():
    """
    Test values written to /etc/selinux/config are lowercase
    """
    mock_file = """
# This file controls the state of SELinux on the system.
# SELINUX= can take one of these three values:
#     enforcing - SELinux security policy is enforced.
#     permissive - SELinux prints warnings instead of enforcing.
#     disabled - No SELinux policy is loaded.
SELINUX=disabled
# SELINUXTYPE= can take one of these three values:
#     targeted - Targeted processes are protected,
#     minimum - Modification of targeted policy. Only selected processes are protected.
#     mls - Multi Level Security protection.
SELINUXTYPE=targeted

"""
    with patch("salt.utils.files.fopen", mock_open(read_data=mock_file)) as m_open:
        selinux.setenforce("Permissive")
        writes = m_open.write_calls()
        assert writes
        for line in writes:
            if line.startswith("SELINUX="):
                assert line == "SELINUX=permissive"


def test_selinux_config_disabled():
    """
    Test values written to /etc/selinux/config are lowercase
    """
    mock_file = """
# This file controls the state of SELinux on the system.
# SELINUX= can take one of these three values:
#     enforcing - SELinux security policy is enforced.
#     permissive - SELinux prints warnings instead of enforcing.
#     disabled - No SELinux policy is loaded.
## SELINUX=disabled
SELINUX=permissive
# SELINUXTYPE= can take one of these three values:
#     targeted - Targeted processes are protected,
#     minimum - Modification of targeted policy. Only selected processes are protected.
#     mls - Multi Level Security protection.
SELINUXTYPE=targeted

"""
    with patch("salt.utils.files.fopen", mock_open(read_data=mock_file)) as m_open:
        selinux.setenforce("Disabled")
        writes = m_open.write_calls()
        assert writes
        for line in writes:
            if line.startswith("SELINUX="):
                assert line == "SELINUX=disabled"



import pytest

import salt.modules.proxy as proxy
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {proxy: {"__grains__": {"os": "Darwin"}}}


def test_get_http_proxy_macos():
    """
    Test to make sure that we correctly get the current proxy info
    on macOS
    """
    mock = MagicMock(
        return_value=(
            "Enabled: Yes\nServer: 192.168.0.1\nPort: 3128\nAuthenticated Proxy"
            " Enabled: 0"
        )
    )
    expected = {"enabled": True, "server": "192.168.0.1", "port": "3128"}

    with patch.dict(proxy.__salt__, {"cmd.run": mock}):
        out = proxy.get_http_proxy()
        mock.assert_called_once_with("networksetup -getwebproxy Ethernet")
        assert expected == out


def test_get_https_proxy_macos():
    """
    Test to make sure that we correctly get the current proxy info
    on macOS
    """
    mock = MagicMock(
        return_value=(
            "Enabled: Yes\nServer: 192.168.0.1\nPort: 3128\nAuthenticated Proxy"
            " Enabled: 0"
        )
    )
    expected = {"enabled": True, "server": "192.168.0.1", "port": "3128"}

    with patch.dict(proxy.__salt__, {"cmd.run": mock}):
        out = proxy.get_https_proxy()
        mock.assert_called_once_with("networksetup -getsecurewebproxy Ethernet")
        assert expected == out


def test_get_ftp_proxy_macos():
    """
    Test to make sure that we correctly get the current proxy info
    on macOS
    """
    mock = MagicMock(
        return_value=(
            "Enabled: Yes\nServer: 192.168.0.1\nPort: 3128\nAuthenticated Proxy"
            " Enabled: 0"
        )
    )
    expected = {"enabled": True, "server": "192.168.0.1", "port": "3128"}

    with patch.dict(proxy.__salt__, {"cmd.run": mock}):
        out = proxy.get_ftp_proxy()
        mock.assert_called_once_with("networksetup -getftpproxy Ethernet")
        assert expected == out


def test_get_http_proxy_macos_none():
    """
    Test to make sure that we correctly return when there's no proxy set
    """
    mock = MagicMock(
        return_value="Enabled: No\nServer:\nPort: 0\nAuthenticated Proxy Enabled: 0"
    )

    with patch.dict(proxy.__salt__, {"cmd.run": mock}):
        out = proxy.get_http_proxy()
        mock.assert_called_once_with("networksetup -getwebproxy Ethernet")
        assert {} == out


def test_set_http_proxy_macos():
    """
    Test to make sure that we correctly set the proxy info
    on macOS
    """
    mock = MagicMock()

    with patch.dict(proxy.__salt__, {"cmd.run": mock}):
        out = proxy.set_http_proxy(
            "192.168.0.1",
            3128,
            "frank",
            "badpassw0rd",
            bypass_hosts=".moo.com,.salt.com",
        )
        mock.assert_called_once_with(
            "networksetup -setwebproxy Ethernet 192.168.0.1 3128 On frank"
            " badpassw0rd"
        )
        assert out


def test_set_https_proxy_macos():
    """
    Test to make sure that we correctly set the proxy info
    on macOS
    """
    mock = MagicMock()

    with patch.dict(proxy.__salt__, {"cmd.run": mock}):
        out = proxy.set_https_proxy(
            "192.168.0.1",
            3128,
            "frank",
            "passw0rd",
            bypass_hosts=".moo.com,.salt.com",
        )
        mock.assert_called_once_with(
            "networksetup -setsecurewebproxy Ethernet 192.168.0.1 3128 On frank"
            " passw0rd"
        )
        assert out


def test_set_ftp_proxy_macos():
    """
    Test to make sure that we correctly set the proxy info
    on macOS
    """
    mock = MagicMock()

    with patch.dict(proxy.__salt__, {"cmd.run": mock}):
        out = proxy.set_ftp_proxy(
            "192.168.0.1",
            3128,
            "frank",
            "badpassw0rd",
            bypass_hosts=".moo.com,.salt.com",
        )
        mock.assert_called_once_with(
            "networksetup -setftpproxy Ethernet 192.168.0.1 3128 On frank"
            " badpassw0rd"
        )
        assert out


def test_get_proxy_win():
    """
    Test to make sure that we correctly get the current proxy info on
    Windows
    """
    result = [
        {
            "vdata": (
                "http=192.168.0.1:3128;https=192.168.0.2:3128;ftp=192.168.0.3:3128"
            )
        },
        {"vdata": 1},
    ]
    mock_reg_read = MagicMock(side_effect=result)
    expected = {
        "enabled": True,
        "http": {"server": "192.168.0.1", "port": "3128"},
        "https": {"server": "192.168.0.2", "port": "3128"},
        "ftp": {"server": "192.168.0.3", "port": "3128"},
    }
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.read_value": mock_reg_read}
    ):
        out = proxy.get_proxy_win()
        assert out == expected
        mock_reg_read.assert_any_call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
        )
        mock_reg_read.assert_any_call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyEnable",
        )


def test_get_http_proxy_windows():
    """
    Test to make sure that we correctly get the current proxy info on
    Windows
    """
    result = {
        "vdata": "http=192.168.0.1:3128;https=192.168.0.2:3128;ftp=192.168.0.3:3128"
    }
    mock = MagicMock(return_value=result)
    expected = {"server": "192.168.0.1", "port": "3128"}
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.read_value": mock}
    ):
        out = proxy.get_http_proxy()
        mock.assert_called_once_with(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
        )
        assert expected == out


def test_get_https_proxy_windows():
    """
    Test to make sure that we correctly get the current proxy info on
    Windows
    """
    result = {
        "vdata": "http=192.168.0.1:3128;https=192.168.0.2:3128;ftp=192.168.0.3:3128"
    }
    mock = MagicMock(return_value=result)
    expected = {"server": "192.168.0.2", "port": "3128"}
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.read_value": mock}
    ):
        out = proxy.get_https_proxy()
        mock.assert_called_once_with(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
        )
        assert expected == out


def test_get_ftp_proxy_windows():
    """
    Test to make sure that we correctly get the current proxy info on
    Windows
    """
    result = {
        "vdata": "http=192.168.0.1:3128;https=192.168.0.2:3128;ftp=192.168.0.3:3128"
    }
    mock = MagicMock(return_value=result)
    expected = {"server": "192.168.0.3", "port": "3128"}
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.read_value": mock}
    ):
        out = proxy.get_ftp_proxy()
        mock.assert_called_once_with(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
        )
        assert expected == out


def test_get_all_proxies_macos_fails():
    mock = MagicMock()
    with patch.dict(proxy.__utils__, {"reg.read_value": mock}):
        out = proxy.get_proxy_win()
        assert not mock.called
        assert out is None


def test_get_all_proxies_windows():
    """
    Test to make sure that we correctly get the current proxy info on
    Windows
    """
    results = [
        {
            "vdata": (
                "http=192.168.0.1:3128;https=192.168.0.2:3128;ftp=192.168.0.3:3128"
            )
        },
        {"vdata": 1},
    ]
    mock = MagicMock(side_effect=results)
    expected = {
        "enabled": True,
        "http": {"server": "192.168.0.1", "port": "3128"},
        "https": {"server": "192.168.0.2", "port": "3128"},
        "ftp": {"server": "192.168.0.3", "port": "3128"},
    }
    calls = [
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyEnable",
        ),
    ]
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.read_value": mock}
    ):
        out = proxy.get_proxy_win()
        mock.assert_has_calls(calls)
        assert expected == out


def test_set_http_proxy_windows():
    """
    Test to make sure that we correctly set the proxy info on Windows
    """
    calls = [
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
            vdata="http=192.168.0.1:3128;",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyEnable",
            vdata=1,
            vtype="REG_DWORD",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyOverride",
            vdata="<local>;.moo.com;.salt.com",
        ),
    ]
    mock_reg = MagicMock()
    mock_cmd = MagicMock()
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.set_value": mock_reg}
    ), patch.dict(proxy.__salt__, {"cmd.run": mock_cmd}):
        out = proxy.set_http_proxy(
            server="192.168.0.1",
            port=3128,
            bypass_hosts=[".moo.com", ".salt.com"],
        )
        mock_reg.assert_has_calls(calls)
        mock_cmd.assert_called_once_with("netsh winhttp import proxy source=ie")
        assert out


def test_set_https_proxy_windows():
    """
    Test to make sure that we correctly set the proxy info on Windows
    """
    calls = [
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
            vdata="https=192.168.0.1:3128;",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyEnable",
            vdata=1,
            vtype="REG_DWORD",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyOverride",
            vdata="<local>;.moo.com;.salt.com",
        ),
    ]
    mock_reg = MagicMock()
    mock_cmd = MagicMock()
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.set_value": mock_reg}
    ), patch.dict(proxy.__salt__, {"cmd.run": mock_cmd}):
        out = proxy.set_https_proxy(
            server="192.168.0.1",
            port=3128,
            bypass_hosts=[".moo.com", ".salt.com"],
        )
        mock_reg.assert_has_calls(calls)
        mock_cmd.assert_called_once_with("netsh winhttp import proxy source=ie")
        assert out


def test_set_ftp_proxy_windows():
    """
    Test to make sure that we correctly set the proxy info on Windows
    """
    calls = [
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
            vdata="ftp=192.168.0.1:3128;",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyEnable",
            vdata=1,
            vtype="REG_DWORD",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyOverride",
            vdata="<local>;.moo.com;.salt.com",
        ),
    ]
    mock_reg = MagicMock()
    mock_cmd = MagicMock()
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.set_value": mock_reg}
    ), patch.dict(proxy.__salt__, {"cmd.run": mock_cmd}):
        out = proxy.set_ftp_proxy(
            server="192.168.0.1",
            port=3128,
            bypass_hosts=[".moo.com", ".salt.com"],
        )
        mock_reg.assert_has_calls(calls)
        mock_cmd.assert_called_once_with("netsh winhttp import proxy source=ie")
        assert out


def test_set_proxy_windows():
    """
    Test to make sure that we correctly set the proxy info on Windows
    """
    calls = [
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
            vdata=(
                "http=192.168.0.1:3128;https=192.168.0.1:3128;ftp=192.168.0.1:3128;"
            ),
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyEnable",
            vdata=1,
            vtype="REG_DWORD",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyOverride",
            vdata="<local>;.moo.com;.salt.com",
        ),
    ]
    mock_reg = MagicMock()
    mock_cmd = MagicMock()
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.set_value": mock_reg}
    ), patch.dict(proxy.__salt__, {"cmd.run": mock_cmd}):
        out = proxy.set_proxy_win(
            server="192.168.0.1",
            port=3128,
            bypass_hosts=[".moo.com", ".salt.com"],
        )
        mock_reg.assert_has_calls(calls)
        mock_cmd.assert_called_once_with("netsh winhttp import proxy source=ie")
        assert out


def test_set_proxy_windows_no_ftp():
    """
    Test to make sure that we correctly set the proxy info on Windows
    """
    calls = [
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyServer",
            vdata="http=192.168.0.1:3128;https=192.168.0.1:3128;",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyEnable",
            vdata=1,
            vtype="REG_DWORD",
        ),
        call(
            hive="HKEY_CURRENT_USER",
            key="SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Internet Settings",
            vname="ProxyOverride",
            vdata="<local>;.moo.com;.salt.com",
        ),
    ]
    mock_reg = MagicMock()
    mock_cmd = MagicMock()
    with patch.dict(proxy.__grains__, {"os": "Windows"}), patch.dict(
        proxy.__utils__, {"reg.set_value": mock_reg}
    ), patch.dict(proxy.__salt__, {"cmd.run": mock_cmd}):
        out = proxy.set_proxy_win(
            server="192.168.0.1",
            port=3128,
            types=["http", "https"],
            bypass_hosts=[".moo.com", ".salt.com"],
        )
        mock_reg.assert_has_calls(calls)
        mock_cmd.assert_called_once_with("netsh winhttp import proxy source=ie")
        assert out



"""
Tests for the salt-run command
"""

import functools
import random
import string
import tempfile

import pytest
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.x509.oid import NameOID

from tests.support.case import ShellCase


def _random_name(prefix=""):
    ret = prefix
    for _ in range(8):
        ret += random.choice(string.ascii_lowercase)
    return ret


def with_random_name(func):
    """
    generate a randomized name for a container
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        name = _random_name(prefix="salt_")
        return func(self, _random_name(prefix="salt-test-"), *args, **kwargs)

    return wrapper


class VenafiTest(ShellCase):
    """
    Test the venafi runner
    """

    @with_random_name
    @pytest.mark.slow_test
    def test_request(self, name):
        cn = "{}.example.com".format(name)

        # Provide python27 compatibility
        if not isinstance(cn, str):
            cn = cn.decode()

        ret = self.run_run_plus(
            fun="venafi.request",
            minion_id=cn,
            dns_name=cn,
            key_password="secretPassword",
            zone="fake",
        )
        cert_output = ret["return"][0]
        assert cert_output is not None, "venafi_certificate not found in `output_value`"

        cert = x509.load_pem_x509_certificate(cert_output.encode(), default_backend())
        assert isinstance(cert, x509.Certificate)
        assert cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME) == [
            x509.NameAttribute(NameOID.COMMON_NAME, cn)
        ]

        pkey_output = ret["return"][1]
        assert pkey_output is not None, "venafi_private key not found in output_value"

        pkey = serialization.load_pem_private_key(
            pkey_output.encode(), password=b"secretPassword", backend=default_backend()
        )

        pkey_public_key_pem = pkey.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        cert_public_key_pem = cert.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        assert pkey_public_key_pem == cert_public_key_pem

    @with_random_name
    @pytest.mark.slow_test
    def test_sign(self, name):

        csr_pem = """-----BEGIN CERTIFICATE REQUEST-----
MIIFbDCCA1QCAQAwgbQxCzAJBgNVBAYTAlVTMQ0wCwYDVQQIDARVdGFoMRIwEAYD
VQQHDAlTYWx0IExha2UxFDASBgNVBAoMC1ZlbmFmaSBJbmMuMRQwEgYDVQQLDAtJ
bnRlZ3JhdGlvbjEnMCUGCSqGSIb3DQEJARYYZW1haWxAdmVuYWZpLmV4YW1wbGUu
Y29tMS0wKwYDVQQDDCR0ZXN0LWNzci0zMjMxMzEzMS52ZW5hZmkuZXhhbXBsZS5j
b20wggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIKAoICAQC4T0bdjq+mF+DABhF+
XWCwOXXUWbPNWa72VVhxoelbyTS0iIeZEe64AvNGykytFdOuT/F9pdkZa+Io07R1
ZMp6Ak8dp2Wjt4c5rayVZus6ZK+0ZwBRJO7if/cqhEpxy8Wz1RMfVLf2AE1u/xZS
QSYY0BTRWGmPqrFJrIGbnyQfvmGVPk3cA0RfdrwYJZXtZ2/4QNrbNCoSoSmqTHzt
NAtZhvT2dPU9U48Prx4b2460x+ck3xA1OdJNXV7n5u53QbxOIcjdGT0lJ62ml70G
5gvEHmdPcg+t5cw/Sm5cfDSUEDtNEXvD4oJXfP98ty6f1cYsZpcrgxRwk9RfGain
hvoweXhZP3NWnU5nRdn2nOfExv+xMeQOyB/rYv98zqzK6LvwKhwI5UB1l/n9KTpg
jgaNCP4x/KAsrPecbHK91oiqGSbPn4wtTYOmPkDxSzATN317u7fE20iqvVAUy/O+
7SCNNKEDPX2NP9LLz0IPK0roQxLiwd2CVyN6kEXuzs/3psptkNRMSlhyeAZdfrOE
CNOp46Pam9f9HGBqzXxxoIlfzLqHHL584kgFlBm7qmivVrgp6zdLPDa+UayXEl2N
O17SnGS8nkOTmfg3cez7lzX/LPLO9X/Y1xKYqx5hoGZhh754K8mzDWCVCYThWgou
yBOYY8uNXiX6ldqzQUHpbxxQgwIDAQABoHIwcAYJKoZIhvcNAQkOMWMwYTBfBgNV
HREEWDBWgilhbHQxLXRlc3QtY3NyLTMyMzEzMTMxLnZlbmFmaS5leGFtcGxlLmNv
bYIpYWx0Mi10ZXN0LWNzci0zMjMxMzEzMS52ZW5hZmkuZXhhbXBsZS5jb20wDQYJ
KoZIhvcNAQELBQADggIBAJd87BIdeh0WWoyQ4IX+ENpNqmm/sLmdfmUB/hj9NpBL
qbr2UTWaSr1jadoZ+mrDxtm1Z0YJDTTIrEWxkBOW5wQ039lYZNe2tfDXSJZwJn7u
2keaXtWQ2SdduK1wOPDO9Hra6WnH7aEq5D1AyoghvPsZwTqZkNynt/A1BZW5C/ha
J9/mwgWfL4qXBGBOhLwKN5GUo3erUkJIdH0TlMqI906D/c/YAuJ86SRdQtBYci6X
bJ7C+OnoiV6USn1HtQE6dfOMeS8voJuixpSIvHZ/Aim6kSAN1Za1f6FQAkyqbF+o
oKTJHDS1CPWikCeLdpPUcOCDIbsiISTsMZkEvIkzZ7dKBIlIugauxw3vaEpk47jN
Wq09r639RbSv/Qs8D6uY66m1IpL4zHm4lTAknrjM/BqihPxc8YiN76ssajvQ4SFT
DHPrDweEVe4KL1ENw8nv4wdkIFKwJTDarV5ZygbETzIhfa2JSBZFTdN+Wmd2Mh5h
OTu+vuHrJF2TO8g1G48EB/KWGt+yvVUpWAanRMwldnFX80NcUlM7GzNn6IXTeE+j
BttIbvAAVJPG8rVCP8u3DdOf+vgm5macj9oLoVP8RBYo/z0E3e+H50nXv3uS6JhN
xlAKgaU6i03jOm5+sww5L2YVMi1eeBN+kx7o94ogpRemC/EUidvl1PUJ6+e7an9V
-----END CERTIFICATE REQUEST-----
        """

        with tempfile.NamedTemporaryFile("w+") as f:
            f.write(csr_pem)
            f.flush()
            csr_path = f.name
            cn = "test-csr-32313131.venafi.example.com"

            # Provide python27 compatibility
            if not isinstance(cn, str):
                cn = cn.decode()

            ret = self.run_run_plus(
                fun="venafi.request", minion_id=cn, csr_path=csr_path, zone="fake"
            )
            cert_output = ret["return"][0]
            assert (
                cert_output is not None
            ), "venafi_certificate not found in `output_value`"

            cert = x509.load_pem_x509_certificate(
                cert_output.encode(), default_backend()
            )
            assert isinstance(cert, x509.Certificate)
            assert cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME) == [
                x509.NameAttribute(NameOID.COMMON_NAME, cn)
            ]



import salt.ext.tornado.gen


def run_loop_in_thread(loop, evt):
    """
    Run the provided loop until an event is set
    """
    loop.make_current()

    @salt.ext.tornado.gen.coroutine
    def stopper():
        while True:
            if evt.is_set():
                loop.stop()
                break
            yield salt.ext.tornado.gen.sleep(0.3)

    loop.add_callback(stopper)
    try:
        loop.start()
    finally:
        loop.close()



"""
.. _saltify-module:

Saltify Module
==============

The Saltify module is designed to install Salt on a remote machine, virtual or
bare metal, using SSH. This module is useful for provisioning machines which
are already installed, but not Salted.

.. versionchanged:: 2018.3.0
    The wake_on_lan capability, and actions destroy, reboot, and query functions were added.

Use of this module requires some configuration in cloud profile and provider
files as described in the
:ref:`Getting Started with Saltify <getting-started-with-saltify>` documentation.
"""

import logging
import time

import salt.client
import salt.config as config
import salt.utils.cloud
from salt._compat import ipaddress
from salt.exceptions import SaltCloudException, SaltCloudSystemExit

log = logging.getLogger(__name__)

try:
    # noinspection PyUnresolvedReferences
    from smbprotocol.exceptions import InternalError as smbSessionError

    HAS_SMB = True
except ImportError:
    HAS_SMB = False

try:
    # noinspection PyUnresolvedReferences
    # noinspection PyUnresolvedReferences
    from requests.exceptions import (
        ConnectionError,
        ConnectTimeout,
        InvalidSchema,
        ProxyError,
        ReadTimeout,
        RetryError,
        SSLError,
    )
    from winrm.exceptions import WinRMTransportError

    HAS_WINRM = True
except ImportError:
    HAS_WINRM = False


def __virtual__():
    """
    Needs no special configuration
    """
    return True


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def avail_locations(call=None):
    """
    This function returns a list of locations available.

    .. code-block:: bash

        salt-cloud --list-locations my-cloud-provider

    [ saltify will always return an empty dictionary ]
    """

    return {}


def avail_images(call=None):
    """
    This function returns a list of images available for this cloud provider.

    .. code-block:: bash

        salt-cloud --list-images saltify

    returns a list of available profiles.

    .. versionadded:: 2018.3.0

    """
    vm_ = get_configured_provider()
    return {"Profiles": [profile for profile in vm_["profiles"]]}


def avail_sizes(call=None):
    """
    This function returns a list of sizes available for this cloud provider.

    .. code-block:: bash

        salt-cloud --list-sizes saltify

    [ saltify always returns an empty dictionary ]
    """
    return {}


def list_nodes(call=None):
    """
    List the nodes which have salt-cloud:driver:saltify grains.

    .. code-block:: bash

        salt-cloud -Q

    returns a list of dictionaries of defined standard fields.

    .. versionadded:: 2018.3.0

    """
    nodes = _list_nodes_full(call)
    return _build_required_items(nodes)


def _build_required_items(nodes):
    ret = {}
    for name, grains in nodes.items():
        if grains:
            private_ips = []
            public_ips = []
            ips = grains["ipv4"] + grains["ipv6"]
            for adrs in ips:
                ip_ = ipaddress.ip_address(adrs)
                if not ip_.is_loopback:
                    if ip_.is_private:
                        private_ips.append(adrs)
                    else:
                        public_ips.append(adrs)

            ret[name] = {
                "id": grains["id"],
                "image": grains["salt-cloud"]["profile"],
                "private_ips": private_ips,
                "public_ips": public_ips,
                "size": "",
                "state": "running",
            }

    return ret


def list_nodes_full(call=None):
    """
    Lists complete information for all nodes.

    .. code-block:: bash

        salt-cloud -F

    returns a list of dictionaries.

    for 'saltify' minions, returns dict of grains (enhanced).

    .. versionadded:: 2018.3.0
    """

    ret = _list_nodes_full(call)

    for (
        key,
        grains,
    ) in ret.items():  # clean up some hyperverbose grains -- everything is too much
        try:
            del (
                grains["cpu_flags"],
                grains["disks"],
                grains["pythonpath"],
                grains["dns"],
                grains["gpus"],
            )
        except KeyError:
            pass  # ignore absence of things we are eliminating
        except TypeError:
            del ret[key]  # eliminate all reference to unexpected (None) values.

    reqs = _build_required_items(ret)

    for name in ret:
        ret[name].update(reqs[name])

    return ret


def _list_nodes_full(call=None):
    """
    List the nodes, ask all 'saltify' minions, return dict of grains.
    """
    with salt.client.LocalClient() as local:
        return local.cmd(
            "salt-cloud:driver:saltify", "grains.items", "", tgt_type="grain"
        )


def list_nodes_select(call=None):
    """
    Return a list of the minions that have salt-cloud grains, with
    select fields.
    """
    return salt.utils.cloud.list_nodes_select(
        list_nodes_full("function"),
        __opts__["query.selection"],
        call,
    )


def show_instance(name, call=None):
    """
    List the a single node, return dict of grains.
    """
    with salt.client.LocalClient() as local:
        ret = local.cmd(name, "grains.items")
        ret.update(_build_required_items(ret))
        return ret


def create(vm_):
    """
    if configuration parameter ``deploy`` is ``True``,

        Provision a single machine, adding its keys to the salt master

    else,

        Test ssh connections to the machine

    Configuration parameters:

    - deploy:  (see above)
    - provider:  name of entry in ``salt/cloud.providers.d/???`` file
    - ssh_host: IP address or DNS name of the new machine
    - ssh_username:  name used to log in to the new machine
    - ssh_password:  password to log in (unless key_filename is used)
    - key_filename:  (optional) SSH private key for passwordless login
    - ssh_port: (default=22) TCP port for SSH connection
    - wake_on_lan_mac:  (optional) hardware (MAC) address for wake on lan
    - wol_sender_node:  (optional) salt minion to send wake on lan command
    - wol_boot_wait:  (default=30) seconds to delay while client boots
    - force_minion_config: (optional) replace the minion configuration files on the new machine

    See also
    :ref:`Miscellaneous Salt Cloud Options <misc-salt-cloud-options>`
    and
    :ref:`Getting Started with Saltify <getting-started-with-saltify>`

    CLI Example:

    .. code-block:: bash

        salt-cloud -p mymachine my_new_id
    """
    deploy_config = config.get_cloud_config_value(
        "deploy", vm_, __opts__, default=False
    )

    # If ssh_host is not set, default to the minion name
    if not config.get_cloud_config_value("ssh_host", vm_, __opts__, default=""):
        vm_["ssh_host"] = vm_["name"]

    if deploy_config:
        wol_mac = config.get_cloud_config_value(
            "wake_on_lan_mac", vm_, __opts__, default=""
        )
        wol_host = config.get_cloud_config_value(
            "wol_sender_node", vm_, __opts__, default=""
        )
        if wol_mac and wol_host:
            good_ping = False
            ssh_host = config.get_cloud_config_value(
                "ssh_host", vm_, __opts__, default=""
            )
            with salt.client.LocalClient() as local:
                if ssh_host:
                    log.info("trying to ping %s", ssh_host)
                    count = "n" if salt.utils.platform.is_windows() else "c"
                    cmd = "ping -{} 1 {}".format(count, ssh_host)
                    good_ping = local.cmd(wol_host, "cmd.retcode", [cmd]) == 0
                if good_ping:
                    log.info("successful ping.")
                else:
                    log.info(
                        "sending wake-on-lan to %s using node %s", wol_mac, wol_host
                    )

                    if isinstance(wol_mac, str):
                        wol_mac = [wol_mac]  # a smart user may have passed more params
                    ret = local.cmd(wol_host, "network.wol", wol_mac)
                    log.info("network.wol returned value %s", ret)
                    if ret and ret[wol_host]:
                        sleep_time = config.get_cloud_config_value(
                            "wol_boot_wait", vm_, __opts__, default=30
                        )
                        if sleep_time > 0.0:
                            log.info("delaying %d seconds for boot", sleep_time)
                            time.sleep(sleep_time)
        log.info("Provisioning existing machine %s", vm_["name"])
        ret = __utils__["cloud.bootstrap"](vm_, __opts__)
    else:
        ret = _verify(vm_)

    return ret


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__, _get_active_provider_name() or "saltify", ()
    )


def _verify(vm_):
    """
    Verify credentials for an existing system
    """
    log.info("Verifying credentials for %s", vm_["name"])

    win_installer = config.get_cloud_config_value("win_installer", vm_, __opts__)

    if win_installer:

        log.debug("Testing Windows authentication method for %s", vm_["name"])

        if not HAS_SMB:
            log.error("smbprotocol library not found")
            return False

        # Test Windows connection
        kwargs = {
            "host": vm_["ssh_host"],
            "username": config.get_cloud_config_value(
                "win_username", vm_, __opts__, default="Administrator"
            ),
            "password": config.get_cloud_config_value(
                "win_password", vm_, __opts__, default=""
            ),
        }

        # Test SMB connection
        try:
            log.debug("Testing SMB protocol for %s", vm_["name"])
            if __utils__["smb.get_conn"](**kwargs) is False:
                return False
        except (smbSessionError) as exc:
            log.error("Exception: %s", exc)
            return False

        # Test WinRM connection
        use_winrm = config.get_cloud_config_value(
            "use_winrm", vm_, __opts__, default=False
        )

        if use_winrm:
            log.debug("WinRM protocol requested for %s", vm_["name"])
            if not HAS_WINRM:
                log.error("WinRM library not found")
                return False

            kwargs["port"] = config.get_cloud_config_value(
                "winrm_port", vm_, __opts__, default=5986
            )
            kwargs["timeout"] = 10

            try:
                log.debug("Testing WinRM protocol for %s", vm_["name"])
                return __utils__["cloud.wait_for_winrm"](**kwargs) is not None
            except (
                ConnectionError,
                ConnectTimeout,
                ReadTimeout,
                SSLError,
                ProxyError,
                RetryError,
                InvalidSchema,
                WinRMTransportError,
            ) as exc:
                log.error("Exception: %s", exc)
                return False

        return True

    else:

        log.debug("Testing SSH authentication method for %s", vm_["name"])

        # Test SSH connection
        kwargs = {
            "host": vm_["ssh_host"],
            "port": config.get_cloud_config_value(
                "ssh_port", vm_, __opts__, default=22
            ),
            "username": config.get_cloud_config_value(
                "ssh_username", vm_, __opts__, default="root"
            ),
            "password": config.get_cloud_config_value(
                "password", vm_, __opts__, search_global=False
            ),
            "key_filename": config.get_cloud_config_value(
                "key_filename",
                vm_,
                __opts__,
                search_global=False,
                default=config.get_cloud_config_value(
                    "ssh_keyfile", vm_, __opts__, search_global=False, default=None
                ),
            ),
            "gateway": vm_.get("gateway", None),
            "maxtries": 1,
        }

        log.debug("Testing SSH protocol for %s", vm_["name"])
        try:
            return __utils__["cloud.wait_for_passwd"](**kwargs) is True
        except SaltCloudException as exc:
            log.error("Exception: %s", exc)
            return False


def destroy(name, call=None):
    """Destroy a node.

    .. versionadded:: 2018.3.0

    Disconnect a minion from the master, and remove its keys.

    Optionally, (if ``remove_config_on_destroy`` is ``True``),
      disables salt-minion from running on the minion, and
      erases the Salt configuration files from it.

    Optionally, (if ``shutdown_on_destroy`` is ``True``),
      orders the minion to halt.

    CLI Example:

    .. code-block:: bash

        salt-cloud --destroy mymachine

    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a, or --action."
        )

    opts = __opts__

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        "salt/cloud/{}/destroying".format(name),
        args={"name": name},
        sock_dir=opts["sock_dir"],
        transport=opts["transport"],
    )

    vm_ = get_configured_provider()
    with salt.client.LocalClient() as local:
        my_info = local.cmd(name, "grains.get", ["salt-cloud"])
        try:
            vm_.update(my_info[name])  # get profile name to get config value
        except (IndexError, TypeError):
            pass
        if config.get_cloud_config_value(
            "remove_config_on_destroy", vm_, opts, default=True
        ):
            ret = local.cmd(
                name,  # prevent generating new keys on restart
                "service.disable",
                ["salt-minion"],
            )
            if ret and ret[name]:
                log.info("disabled salt-minion service on %s", name)
            ret = local.cmd(name, "config.get", ["conf_file"])
            if ret and ret[name]:
                confile = ret[name]
                ret = local.cmd(name, "file.remove", [confile])
                if ret and ret[name]:
                    log.info("removed minion %s configuration file %s", name, confile)
            ret = local.cmd(name, "config.get", ["pki_dir"])
            if ret and ret[name]:
                pki_dir = ret[name]
                ret = local.cmd(name, "file.remove", [pki_dir])
                if ret and ret[name]:
                    log.info("removed minion %s key files in %s", name, pki_dir)

        if config.get_cloud_config_value(
            "shutdown_on_destroy", vm_, opts, default=False
        ):
            ret = local.cmd(name, "system.shutdown")
            if ret and ret[name]:
                log.info("system.shutdown for minion %s successful", name)

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        "salt/cloud/{}/destroyed".format(name),
        args={"name": name},
        sock_dir=opts["sock_dir"],
        transport=opts["transport"],
    )

    return {"Destroyed": "{} was destroyed.".format(name)}


def reboot(name, call=None):
    """
    Reboot a saltify minion.

    .. versionadded:: 2018.3.0

    name
        The name of the VM to reboot.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a reboot vm_name
    """

    if call != "action":
        raise SaltCloudException(
            "The reboot action must be called with -a or --action."
        )

    with salt.client.LocalClient() as local:
        return local.cmd(name, "system.reboot")



import os
import shutil
import time

import pytest
from pytestshellutils.utils import ports
from saltfactories.daemons import master
from saltfactories.daemons.container import SaltDaemon, SaltMinion
from saltfactories.utils import random_string

from salt.version import SaltVersionsInfo, __version__

pytestmark = [pytest.mark.skip_if_binaries_missing("docker")]


class ContainerMaster(SaltDaemon, master.SaltMaster):
    """
    Containerized salt master that has no check events
    """

    def get_display_name(self):
        return master.SaltMaster.get_display_name(self)

    def get_check_events(self):
        return []


class ContainerMinion(SaltMinion):
    """
    Containerized salt minion that has no check events
    """

    def get_check_events(self):
        return []


# ---------------------- Previous Version Setup ----------------------


@pytest.fixture
def prev_version():
    return str(SaltVersionsInfo.previous_release().info[0])


@pytest.fixture
def curr_version():
    return str(SaltVersionsInfo.current_release().info[0])


@pytest.fixture
def prev_master_id():
    return random_string("master-performance-prev-", uppercase=False)


@pytest.fixture
def prev_master(
    request,
    salt_factories,
    host_docker_network_ip_address,
    network,
    prev_version,
    docker_client,
    prev_master_id,
):
    root_dir = salt_factories.get_root_dir_for_daemon(prev_master_id)
    conf_dir = root_dir / "conf"
    conf_dir.mkdir(exist_ok=True)

    config_defaults = {
        "root_dir": str(root_dir),
        "transport": request.config.getoption("--transport"),
        "user": False,
    }
    publish_port = ports.get_unused_localhost_port()
    ret_port = ports.get_unused_localhost_port()
    config_overrides = {
        "interface": "0.0.0.0",
        "publish_port": publish_port,
        "ret_port": ret_port,
        "log_level_logfile": "quiet",
        "pytest-master": {
            "log": {"host": host_docker_network_ip_address},
        },
    }

    factory = salt_factories.salt_master_daemon(
        prev_master_id,
        defaults=config_defaults,
        overrides=config_overrides,
        factory_class=ContainerMaster,
        image="ghcr.io/saltstack/salt-ci-containers/salt:{}".format(prev_version),
        base_script_args=["--log-level=debug"],
        container_run_kwargs={
            "network": network,
            "hostname": prev_master_id,
        },
        docker_client=docker_client,
        name=prev_master_id,
        start_timeout=120,
        max_start_attempts=1,
        skip_if_docker_client_not_connectable=True,
    )
    with factory.started():
        yield factory


@pytest.fixture
def prev_salt_cli(prev_master):
    return prev_master.salt_cli()


@pytest.fixture
def prev_salt_key_cli(prev_master):
    return prev_master.salt_key_cli()


@pytest.fixture
def prev_salt_run_cli(prev_master):
    return prev_master.salt_run_cli()


@pytest.fixture
def prev_minion_id():
    return random_string(
        "minion-performance-prev-",
        uppercase=False,
    )


@pytest.fixture
def prev_minion(
    prev_minion_id,
    prev_master,
    docker_client,
    prev_version,
    host_docker_network_ip_address,
    network,
    prev_master_id,
):
    config_overrides = {
        "master": prev_master_id,
        "user": False,
        "pytest-minion": {"log": {"host": host_docker_network_ip_address}},
    }
    factory = prev_master.salt_minion_daemon(
        prev_minion_id,
        overrides=config_overrides,
        factory_class=ContainerMinion,
        # SaltMinion kwargs
        name=prev_minion_id,
        image="ghcr.io/saltstack/salt-ci-containers/salt:{}".format(prev_version),
        docker_client=docker_client,
        start_timeout=120,
        pull_before_start=False,
        skip_if_docker_client_not_connectable=True,
        container_run_kwargs={
            "network": network,
            "hostname": prev_minion_id,
        },
        max_start_attempts=1,
    )
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, prev_master, factory.id
    )
    with factory.started():
        yield factory


@pytest.fixture
def prev_sls(sls_contents, state_tree, tmp_path):
    sls_name = "prev"
    location = tmp_path / "prev" / "testfile"
    location.parent.mkdir()
    with pytest.helpers.temp_file(
        "{}.sls".format(sls_name), sls_contents.format(path=str(location)), state_tree
    ):
        yield sls_name


# ---------------------- Current Version Setup ----------------------


def _install_local_salt(factory):
    factory.run("pip install /saltcode")


@pytest.fixture
def curr_master_id():
    return random_string("master-performance-", uppercase=False)


@pytest.fixture
def curr_master(
    request,
    salt_factories,
    host_docker_network_ip_address,
    network,
    docker_client,
    curr_master_id,
):
    root_dir = salt_factories.get_root_dir_for_daemon(curr_master_id)
    conf_dir = root_dir / "conf"
    conf_dir.mkdir(exist_ok=True)

    config_defaults = {
        "root_dir": str(root_dir),
        "transport": request.config.getoption("--transport"),
        "user": False,
    }
    publish_port = ports.get_unused_localhost_port()
    ret_port = ports.get_unused_localhost_port()
    config_overrides = {
        "interface": "0.0.0.0",
        "publish_port": publish_port,
        "ret_port": ret_port,
        "log_level_logfile": "quiet",
        "pytest-master": {
            "log": {"host": host_docker_network_ip_address},
        },
    }

    factory = salt_factories.salt_master_daemon(
        curr_master_id,
        defaults=config_defaults,
        overrides=config_overrides,
        factory_class=ContainerMaster,
        image="ghcr.io/saltstack/salt-ci-containers/salt:current",
        base_script_args=["--log-level=debug"],
        container_run_kwargs={
            "network": network,
            "hostname": curr_master_id,
            # Bind the current code to a directory for pip installing
            "volumes": {
                os.environ["REPO_ROOT_DIR"]: {"bind": "/saltcode", "mode": "z"}
            },
        },
        docker_client=docker_client,
        name=curr_master_id,
        start_timeout=120,
        max_start_attempts=1,
        skip_if_docker_client_not_connectable=True,
    )

    factory.before_start(_install_local_salt, factory)
    with factory.started():
        yield factory


@pytest.fixture
def curr_salt_cli(curr_master):
    return curr_master.salt_cli()


@pytest.fixture
def curr_salt_run_cli(curr_master):
    return curr_master.salt_run_cli()


@pytest.fixture
def curr_salt_key_cli(curr_master):
    return curr_master.salt_key_cli()


@pytest.fixture
def curr_minion_id():
    return random_string(
        "minion-performance-curr-",
        uppercase=False,
    )


@pytest.fixture
def curr_minion(
    curr_minion_id,
    curr_master,
    docker_client,
    host_docker_network_ip_address,
    network,
    curr_master_id,
):
    config_overrides = {
        "master": curr_master_id,
        "user": False,
        "pytest-minion": {"log": {"host": host_docker_network_ip_address}},
    }
    factory = curr_master.salt_minion_daemon(
        curr_minion_id,
        overrides=config_overrides,
        factory_class=ContainerMinion,
        # SaltMinion kwargs
        name=curr_minion_id,
        image="ghcr.io/saltstack/salt-ci-containers/salt:current",
        docker_client=docker_client,
        start_timeout=120,
        pull_before_start=False,
        skip_if_docker_client_not_connectable=True,
        container_run_kwargs={
            "network": network,
            "hostname": curr_minion_id,
            # Bind the current code to a directory for pip installing
            "volumes": {
                os.environ["REPO_ROOT_DIR"]: {"bind": "/saltcode", "mode": "z"}
            },
        },
        max_start_attempts=1,
    )
    factory.before_start(_install_local_salt, factory)
    factory.after_terminate(
        pytest.helpers.remove_stale_minion_key, curr_master, factory.id
    )
    with factory.started():
        yield factory


@pytest.fixture
def curr_sls(sls_contents, state_tree, tmp_path):
    sls_name = "curr"
    location = tmp_path / "curr" / "testfile"
    location.parent.mkdir()
    with pytest.helpers.temp_file(
        "{}.sls".format(sls_name), sls_contents.format(path=str(location)), state_tree
    ):
        yield sls_name


def _wait_for_stdout(expected, func, *args, timeout=120, **kwargs):
    start = time.time()
    while time.time() < start + timeout:
        ret = func(*args, **kwargs)
        if ret and ret.stdout and expected in ret.stdout:
            break
        time.sleep(1)
    else:
        pytest.skip(
            f"Skipping test, one or more daemons failed to start: {expected} not found in {ret}"
        )


@pytest.mark.flaky(max_runs=4)
def test_performance(
    prev_salt_cli,
    prev_minion,
    prev_salt_run_cli,
    prev_salt_key_cli,
    prev_version,
    prev_master,
    state_tree,
    curr_salt_cli,
    curr_master,
    curr_salt_run_cli,
    curr_salt_key_cli,
    curr_minion,
    prev_sls,
    curr_sls,
    curr_version,
):
    # Copy all of the needed files to both master file roots directories
    subdir = random_string("performance-")
    shutil.copytree(
        state_tree, os.path.join(curr_master.config["file_roots"]["base"][0], subdir)
    )
    shutil.copytree(
        state_tree, os.path.join(prev_master.config["file_roots"]["base"][0], subdir)
    )

    # Wait for the old master and minion to start
    _wait_for_stdout(
        prev_version, prev_master.run, *prev_salt_run_cli.cmdline("--version")
    )
    salt_key_cmd = [
        comp
        for comp in prev_salt_key_cli.cmdline("-Ay")
        if not comp.startswith("--log-level")
    ]
    _wait_for_stdout(prev_minion.id, prev_master.run, *salt_key_cmd)
    _wait_for_stdout(
        "Salt: {}".format(prev_version),
        prev_master.run,
        *prev_salt_cli.cmdline("test.versions", minion_tgt=prev_minion.id),
    )

    # Wait for the new master and minion to start
    _wait_for_stdout(
        curr_version, curr_master.run, *curr_salt_run_cli.cmdline("--version")
    )
    curr_key_cmd = [
        comp
        for comp in curr_salt_key_cli.cmdline("-Ay")
        if not comp.startswith("--log-level")
    ]
    _wait_for_stdout(curr_minion.id, curr_master.run, *curr_key_cmd)
    _wait_for_stdout(
        "Salt: {}".format(curr_version),
        curr_master.run,
        *curr_salt_cli.cmdline("test.versions", minion_tgt=curr_minion.id),
    )

    # Let's now apply the states
    applies = os.environ.get("SALT_PERFORMANCE_TEST_APPLIES", 3)

    def _gather_durations(ret, minion_id):
        """
        Get the total duration for the state run.

        We skip if anything fails here.  We aren't testing state success, just performance.
        """
        if isinstance(ret.data, dict) and isinstance(
            ret.data.get(minion_id, None), dict
        ):
            duration = 0
            for _, state_ret in ret.data[minion_id].items():
                try:
                    duration += state_ret["duration"]
                except KeyError:
                    break
            else:
                return duration
        pytest.skip("Something went wrong with the states, skipping.")

    prev_duration = 0
    curr_duration = 0

    for _ in range(applies):
        prev_state_ret = prev_master.run(
            *prev_salt_cli.cmdline(
                "state.apply", f"{subdir}.{prev_sls}", minion_tgt=prev_minion.id
            )
        )
        prev_duration += _gather_durations(prev_state_ret, prev_minion.id)

    for _ in range(applies):
        curr_state_ret = curr_master.run(
            *curr_salt_cli.cmdline(
                "state.apply", f"{subdir}.{curr_sls}", minion_tgt=curr_minion.id
            )
        )
        curr_duration += _gather_durations(curr_state_ret, curr_minion.id)

    # We account for network slowness, etc... here.
    # There is a hard balance here as far as a threshold.
    # We need to make sure there are no drastic slowdowns,
    # but also take into account system and network nuances.
    # In theory we could set a hard cap for the duration,
    # something like 500 ms and only run the current version,
    # but we will see if this ever becomes too flaky
    assert curr_duration <= 1.25 * prev_duration



"""
Tests for salt.utils.jinja
"""

import ast
import itertools
import os
import pprint
import re

import pytest
from jinja2 import DictLoader, Environment, exceptions

import salt.loader

# dateutils is needed so that the strftime jinja filter is loaded
import salt.modules.match as match
import salt.utils.dateutils  # pylint: disable=unused-import
import salt.utils.files
import salt.utils.json
import salt.utils.stringutils
import salt.utils.yaml
from salt.exceptions import SaltRenderError
from salt.utils.decorators.jinja import JinjaFilter
from salt.utils.jinja import SerializerExtension, ensure_sequence_filter
from salt.utils.odict import OrderedDict
from salt.utils.templates import render_jinja_tmpl

try:
    import timelib  # pylint: disable=W0611

    HAS_TIMELIB = True
except ImportError:
    HAS_TIMELIB = False


@pytest.fixture
def minion_opts(tmp_path, minion_opts):
    minion_opts.update(
        {
            "cachedir": str(tmp_path / "jinja-template-cache"),
            "file_buffer_size": 1048576,
            "file_client": "local",
            "file_ignore_regex": None,
            "file_ignore_glob": None,
            "file_roots": {"test": [str(tmp_path / "templates")]},
            "pillar_roots": {"test": [str(tmp_path / "templates")]},
            "fileserver_backend": ["roots"],
            "hash_type": "md5",
            "extension_modules": os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "extmods"
            ),
        }
    )
    return minion_opts


@pytest.fixture()
def configure_loader_modules(minion_opts):
    return {match: {"__opts__": minion_opts}}


@pytest.fixture
def local_salt():
    return {}


def test_regex_escape():
    dataset = "foo?:.*/\\bar"
    env = Environment(extensions=[SerializerExtension])
    env.filters.update(JinjaFilter.salt_jinja_filters)
    rendered = env.from_string("{{ dataset|regex_escape }}").render(dataset=dataset)
    assert rendered == re.escape(dataset)


def test_unique_string():
    dataset = "foo"
    unique = set(dataset)
    env = Environment(extensions=[SerializerExtension])
    env.filters.update(JinjaFilter.salt_jinja_filters)
    rendered = (
        env.from_string("{{ dataset|unique }}")
        .render(dataset=dataset)
        .strip("'{}")
        .split("', '")
    )
    assert sorted(rendered) == sorted(list(unique))


def test_unique_tuple():
    dataset = ("foo", "foo", "bar")
    unique = set(dataset)
    env = Environment(extensions=[SerializerExtension])
    env.filters.update(JinjaFilter.salt_jinja_filters)
    rendered = (
        env.from_string("{{ dataset|unique }}")
        .render(dataset=dataset)
        .strip("'{}")
        .split("', '")
    )
    assert sorted(rendered) == sorted(list(unique))


def test_unique_list():
    dataset = ["foo", "foo", "bar"]
    unique = ["foo", "bar"]
    env = Environment(extensions=[SerializerExtension])
    env.filters.update(JinjaFilter.salt_jinja_filters)
    rendered = (
        env.from_string("{{ dataset|unique }}")
        .render(dataset=dataset)
        .strip("'[]")
        .split("', '")
    )
    assert rendered == unique


def test_serialize_json():
    dataset = {"foo": True, "bar": 42, "baz": [1, 2, 3], "qux": 2.0}
    env = Environment(extensions=[SerializerExtension])
    rendered = env.from_string("{{ dataset|json }}").render(dataset=dataset)
    assert dataset == salt.utils.json.loads(rendered)


def test_serialize_yaml():
    dataset = {
        "foo": True,
        "bar": 42,
        "baz": [1, 2, 3],
        "qux": 2.0,
        "spam": OrderedDict([("foo", OrderedDict([("bar", "baz"), ("qux", 42)]))]),
    }
    env = Environment(extensions=[SerializerExtension])
    rendered = env.from_string("{{ dataset|yaml }}").render(dataset=dataset)
    assert dataset == salt.utils.yaml.safe_load(rendered)


def test_serialize_yaml_str():
    dataset = "str value"
    env = Environment(extensions=[SerializerExtension])
    rendered = env.from_string("{{ dataset|yaml }}").render(dataset=dataset)
    assert dataset == rendered


def test_serialize_yaml_unicode():
    dataset = "str value"
    env = Environment(extensions=[SerializerExtension])
    rendered = env.from_string("{{ dataset|yaml }}").render(dataset=dataset)
    assert "str value" == rendered


def test_serialize_python():
    dataset = {"foo": True, "bar": 42, "baz": [1, 2, 3], "qux": 2.0}
    env = Environment(extensions=[SerializerExtension])
    rendered = env.from_string("{{ dataset|python }}").render(dataset=dataset)
    assert rendered == pprint.pformat(dataset)


def test_load_yaml():
    env = Environment(extensions=[SerializerExtension])
    rendered = env.from_string(
        '{% set document = "{foo: it works}"|load_yaml %}{{ document.foo }}'
    ).render()
    assert rendered == "it works"

    rendered = env.from_string(
        "{% set document = document|load_yaml %}{{ document.foo }}"
    ).render(document="{foo: it works}")
    assert rendered == "it works"

    with pytest.raises((TypeError, exceptions.TemplateRuntimeError)):
        env.from_string(
            "{% set document = document|load_yaml %}{{ document.foo }}"
        ).render(document={"foo": "it works"})


def test_load_tag():
    env = Environment(extensions=[SerializerExtension])

    source = (
        "{{ bar }}, "
        + "{% load_yaml as docu %}{foo: it works, {{ bar }}: baz}{% endload %}"
        + "{{ docu.foo }}"
    )

    rendered = env.from_string(source).render(bar="barred")
    assert rendered == "barred, it works"

    source = (
        '{{ bar }}, {% load_json as docu %}{"foo": "it works", "{{ bar }}":'
        ' "baz"}{% endload %}' + "{{ docu.foo }}"
    )

    rendered = env.from_string(source).render(bar="barred")
    assert rendered == "barred, it works"

    with pytest.raises(exceptions.TemplateSyntaxError):
        env.from_string(
            "{% load_yamle as document %}{foo, bar: it works}{% endload %}"
        ).render()

    with pytest.raises(exceptions.TemplateRuntimeError):
        env.from_string(
            "{% load_json as document %}{foo, bar: it works}{% endload %}"
        ).render()


def test_load_json():
    env = Environment(extensions=[SerializerExtension])
    rendered = env.from_string(
        """{% set document = '{"foo": "it works"}'|load_json %}{{ document.foo }}"""
    ).render()
    assert rendered == "it works"

    rendered = env.from_string(
        "{% set document = document|load_json %}{{ document.foo }}"
    ).render(document='{"foo": "it works"}')
    assert rendered == "it works"

    # bad quotes
    with pytest.raises(exceptions.TemplateRuntimeError):
        env.from_string("{{ document|load_json }}").render(
            document="{'foo': 'it works'}"
        )

    # not a string
    with pytest.raises(exceptions.TemplateRuntimeError):
        env.from_string("{{ document|load_json }}").render(document={"foo": "it works"})


def test_load_yaml_template():
    loader = DictLoader({"foo": '{bar: "my god is blue", foo: [1, 2, 3]}'})
    env = Environment(extensions=[SerializerExtension], loader=loader)
    rendered = env.from_string('{% import_yaml "foo" as doc %}{{ doc.bar }}').render()
    assert rendered == "my god is blue"

    with pytest.raises(exceptions.TemplateNotFound):
        env.from_string('{% import_yaml "does not exists" as doc %}').render()


def test_load_json_template():
    loader = DictLoader({"foo": '{"bar": "my god is blue", "foo": [1, 2, 3]}'})
    env = Environment(extensions=[SerializerExtension], loader=loader)
    rendered = env.from_string('{% import_json "foo" as doc %}{{ doc.bar }}').render()
    assert rendered == "my god is blue"

    with pytest.raises(exceptions.TemplateNotFound):
        env.from_string('{% import_json "does not exists" as doc %}').render()


def test_load_text_template():
    loader = DictLoader({"foo": "Foo!"})
    env = Environment(extensions=[SerializerExtension], loader=loader)

    rendered = env.from_string('{% import_text "foo" as doc %}{{ doc }}').render()
    assert rendered == "Foo!"

    with pytest.raises(exceptions.TemplateNotFound):
        env.from_string('{% import_text "does not exists" as doc %}').render()


def test_profile():
    env = Environment(extensions=[SerializerExtension])

    source = (
        "{%- profile as 'profile test' %}"
        "{% set var = 'val' %}"
        "{%- endprofile %}"
        "{{ var }}"
    )

    rendered = env.from_string(source).render()
    assert rendered == "val"


def test_catalog():
    loader = DictLoader(
        {
            "doc1": '{bar: "my god is blue"}',
            "doc2": '{% import_yaml "doc1" as local2 %} never exported',
            "doc3": (
                '{% load_yaml as local3 %}{"foo": "it works"}{% endload %} me'
                " neither"
            ),
            "main1": '{% from "doc2" import local2 %}{{ local2.bar }}',
            "main2": '{% from "doc3" import local3 %}{{ local3.foo }}',
            "main3": """
            {% import "doc2" as imported2 %}
            {% import "doc3" as imported3 %}
            {{ imported2.local2.bar }}
        """,
            "main4": """
            {% import "doc2" as imported2 %}
            {% import "doc3" as imported3 %}
            {{ imported3.local3.foo }}
        """,
            "main5": """
            {% from "doc2" import local2 as imported2 %}
            {% from "doc3" import local3 as imported3 %}
            {{ imported2.bar }}
        """,
            "main6": """
            {% from "doc2" import local2 as imported2 %}
            {% from "doc3" import local3 as imported3 %}
            {{ imported3.foo }}
        """,
        }
    )

    env = Environment(extensions=[SerializerExtension], loader=loader)
    rendered = env.get_template("main1").render()
    assert rendered == "my god is blue"

    rendered = env.get_template("main2").render()
    assert rendered == "it works"

    rendered = env.get_template("main3").render().strip()
    assert rendered == "my god is blue"

    rendered = env.get_template("main4").render().strip()
    assert rendered == "it works"

    rendered = env.get_template("main5").render().strip()
    assert rendered == "my god is blue"

    rendered = env.get_template("main6").render().strip()
    assert rendered == "it works"


def test_nested_structures():
    env = Environment(extensions=[SerializerExtension])
    rendered = env.from_string("{{ data }}").render(data="foo")
    assert rendered == "foo"

    data = OrderedDict([("foo", OrderedDict([("bar", "baz"), ("qux", 42)]))])

    rendered = env.from_string("{{ data }}").render(data=data)
    assert rendered == "{'foo': {'bar': 'baz', 'qux': 42}}"

    rendered = env.from_string("{{ data }}").render(
        data=[
            OrderedDict(
                foo="bar",
            ),
            OrderedDict(
                baz=42,
            ),
        ]
    )
    assert rendered == "[{'foo': 'bar'}, {'baz': 42}]"


def test_set_dict_key_value(minion_opts, local_salt):
    """
    Test the `set_dict_key_value` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ {} | set_dict_key_value('foo:bar:baz', 42) }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "{'foo': {'bar': {'baz': 42}}}"

    rendered = render_jinja_tmpl(
        "{{ {} | set_dict_key_value('foo.bar.baz', 42, delimiter='.') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "{'foo': {'bar': {'baz': 42}}}"


def test_update_dict_key_value(minion_opts, local_salt):
    """
    Test the `update_dict_key_value` Jinja filter.
    """
    # Use OrderedDicts to avoid random key-order-switches in the rendered string.
    expected = OrderedDict(
        [("bar", OrderedDict([("baz", OrderedDict([("qux", 1), ("quux", 3)]))]))]
    )
    dataset = OrderedDict([("bar", OrderedDict([("baz", OrderedDict([("qux", 1)]))]))])
    dataset_exp = OrderedDict([("quux", 3)])
    rendered = render_jinja_tmpl(
        "{{ foo | update_dict_key_value('bar:baz', exp) }}",
        dict(
            foo=dataset,
            exp=dataset_exp,
            opts=minion_opts,
            saltenv="test",
            salt=local_salt,
        ),
    )
    assert rendered == "{'bar': {'baz': {'qux': 1, 'quux': 3}}}"

    # Test incorrect usage
    for update_with in [42, "foo", [42]]:
        template = "{{ {} | update_dict_key_value('bar:baz', update_with) }}"
        expected = r"Cannot update {} with a {}.".format(type({}), type(update_with))
        with pytest.raises(SaltRenderError, match=expected):
            render_jinja_tmpl(
                template,
                dict(
                    update_with=update_with,
                    opts=minion_opts,
                    saltenv="test",
                    salt=local_salt,
                ),
            )


def test_append_dict_key_value(minion_opts, local_salt):
    """
    Test the `append_dict_key_value` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ {} | append_dict_key_value('foo:bar:baz', 42) }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "{'foo': {'bar': {'baz': [42]}}}"

    rendered = render_jinja_tmpl(
        "{{ foo | append_dict_key_value('bar:baz', 42) }}",
        dict(
            foo={"bar": {"baz": [1, 2]}},
            opts=minion_opts,
            saltenv="test",
            salt=local_salt,
        ),
    )
    assert rendered == "{'bar': {'baz': [1, 2, 42]}}"


def test_extend_dict_key_value(minion_opts, local_salt):
    """
    Test the `extend_dict_key_value` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ {} | extend_dict_key_value('foo:bar:baz', [42]) }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "{'foo': {'bar': {'baz': [42]}}}"

    rendered = render_jinja_tmpl(
        "{{ foo | extend_dict_key_value('bar:baz', [42, 43]) }}",
        dict(
            foo={"bar": {"baz": [1, 2]}},
            opts=minion_opts,
            saltenv="test",
            salt=local_salt,
        ),
    )
    assert rendered == "{'bar': {'baz': [1, 2, 42, 43]}}"
    # Edge cases
    rendered = render_jinja_tmpl(
        "{{ {} | extend_dict_key_value('foo:bar:baz', 'quux') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "{'foo': {'bar': {'baz': ['q', 'u', 'u', 'x']}}}"
    # Beware! When supplying a dict, the list gets extended with the dict coerced to a list,
    # which will only contain the keys of the dict.
    rendered = render_jinja_tmpl(
        "{{ {} | extend_dict_key_value('foo:bar:baz', {'foo': 'bar'}) }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "{'foo': {'bar': {'baz': ['foo']}}}"

    # Test incorrect usage
    template = "{{ {} | extend_dict_key_value('bar:baz', 42) }}"
    expected = r"Cannot extend {} with a {}.".format(type([]), int)
    with pytest.raises(SaltRenderError, match=expected):
        render_jinja_tmpl(
            template, dict(opts=minion_opts, saltenv="test", salt=local_salt)
        )


def test_sequence():
    env = Environment()
    env.filters["sequence"] = ensure_sequence_filter

    rendered = env.from_string("{{ data | sequence | length }}").render(data="foo")
    assert rendered == "1"

    rendered = env.from_string("{{ data | sequence | length }}").render(
        data=["foo", "bar"]
    )
    assert rendered == "2"

    rendered = env.from_string("{{ data | sequence | length }}").render(
        data=("foo", "bar")
    )
    assert rendered == "2"

    rendered = env.from_string("{{ data | sequence | length }}").render(
        data={"foo", "bar"}
    )
    assert rendered == "2"

    rendered = env.from_string("{{ data | sequence | length }}").render(
        data={"foo": "bar"}
    )
    assert rendered == "1"


def test_camel_to_snake_case(minion_opts, local_salt):
    """
    Test the `to_snake_case` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'abcdEfghhIjkLmnoP' | to_snake_case }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "abcd_efghh_ijk_lmno_p"


def test_snake_to_camel_case(minion_opts, local_salt):
    """
    Test the `to_camelcase` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'the_fox_jumped_over_the_lazy_dog' | to_camelcase }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "theFoxJumpedOverTheLazyDog"

    rendered = render_jinja_tmpl(
        "{{ 'the_fox_jumped_over_the_lazy_dog' | to_camelcase(uppercamel=True) }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "TheFoxJumpedOverTheLazyDog"


def test_is_ip(minion_opts, local_salt):
    """
    Test the `is_ip` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1' | is_ip }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "True"

    rendered = render_jinja_tmpl(
        "{{ 'FE80::' | is_ip }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "True"

    rendered = render_jinja_tmpl(
        "{{ 'random' | is_ip }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "False"


def test_is_ipv4(minion_opts, local_salt):
    """
    Test the `is_ipv4` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1' | is_ipv4 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "True"

    rendered = render_jinja_tmpl(
        "{{ 'FE80::' | is_ipv4 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "False"

    rendered = render_jinja_tmpl(
        "{{ 'random' | is_ipv4 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "False"


def test_is_ipv6(minion_opts):
    """
    Test the `is_ipv6` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1' | is_ipv6 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "False"

    rendered = render_jinja_tmpl(
        "{{ 'fe80::20d:b9ff:fe01:ea8%eth0' | is_ipv6 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "True"

    rendered = render_jinja_tmpl(
        "{{ 'FE80::' | is_ipv6 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "True"

    rendered = render_jinja_tmpl(
        "{{ 'random' | is_ipv6 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "False"


def test_ipaddr(minion_opts, local_salt):
    """
    Test the `ipaddr` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ '::' | ipaddr }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "::"

    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1' | ipaddr }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "192.168.0.1"

    # provides a list with valid IP addresses only
    rendered = render_jinja_tmpl(
        "{{ ['192.168.0.1', '172.17.17.1', 'foo', 'bar', '::'] | ipaddr | join(',"
        " ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "192.168.0.1, 172.17.17.1, ::"

    # return only multicast addresses
    rendered = render_jinja_tmpl(
        "{{ ['224.0.0.1', 'FF01::1', '::'] | ipaddr(options='multicast') | join(',"
        " ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "224.0.0.1, ff01::1"


def test_ipv4(minion_opts, local_salt):
    """
    Test the `ipv4` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1' | ipv4 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "192.168.0.1"

    rendered = render_jinja_tmpl(
        "{{ ['192.168.0.1', '172.17.17.1'] | ipv4 | join(', ')}}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "192.168.0.1, 172.17.17.1"

    rendered = render_jinja_tmpl(
        "{{ 'fe80::' | ipv4 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "None"

    rendered = render_jinja_tmpl(
        "{{ 'random' | ipv4 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "None"

    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1' | ipv4(options='lo') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "None"

    rendered = render_jinja_tmpl(
        "{{ '127.0.0.1' | ipv4(options='lo') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "127.0.0.1"


def test_ipv6(minion_opts, local_salt):
    """
    Test the `ipv6` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1' | ipv6 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "None"

    rendered = render_jinja_tmpl(
        "{{ 'random' | ipv6 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "None"

    # returns the standard format value
    rendered = render_jinja_tmpl(
        "{{ 'FE80:0:0::0' | ipv6 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "fe80::"

    # fe80:: is link local therefore will be returned
    rendered = render_jinja_tmpl(
        "{{ 'fe80::' | ipv6(options='ll') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "fe80::"

    # fe80:: is not loopback
    rendered = render_jinja_tmpl(
        "{{ 'fe80::' | ipv6(options='lo') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "None"

    # returns only IPv6 addresses in the list
    rendered = render_jinja_tmpl(
        "{{ ['fe80::', '192.168.0.1'] | ipv6 | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "fe80::"

    rendered = render_jinja_tmpl(
        "{{ ['fe80::', '::'] | ipv6 | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "fe80::, ::"


def test_ipwrap(minion_opts, local_salt):
    """
    Test the `ipwrap` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1' | ipwrap }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "192.168.0.1"

    rendered = render_jinja_tmpl(
        "{{ 'random' | ipwrap }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "random"

    # returns the standard format value
    rendered = render_jinja_tmpl(
        "{{ 'FE80:0:0::0' | ipwrap }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "[fe80::]"

    rendered = render_jinja_tmpl(
        "{{ ['fe80::', '::'] | ipwrap | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "[fe80::], [::]"

    rendered = render_jinja_tmpl(
        "{{ ['fe80::', 'ham', 'spam', '2001:db8::1', 'eggs', '::'] | ipwrap | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "[fe80::], ham, spam, [2001:db8::1], eggs, [::]"

    rendered = render_jinja_tmpl(
        "{{ ('fe80::', 'ham', 'spam', '2001:db8::1', 'eggs', '::') | ipwrap | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "[fe80::], ham, spam, [2001:db8::1], eggs, [::]"


def test_network_hosts(minion_opts, local_salt):
    """
    Test the `network_hosts` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1/30' | network_hosts | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "192.168.0.1, 192.168.0.2"


def test_network_size(minion_opts, local_salt):
    """
    Test the `network_size` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1' | network_size }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "1"

    rendered = render_jinja_tmpl(
        "{{ '192.168.0.1/8' | network_size }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "16777216"


@pytest.mark.requires_network
@pytest.mark.parametrize("backend", ["requests", "tornado", "urllib2"])
def test_http_query(minion_opts, local_salt, backend, httpserver):
    """
    Test the `http_query` Jinja filter.
    """
    urls = (
        # These cannot be HTTPS urls since urllib2 chokes on those
        "http://saltproject.io",
        "http://google.com",
        "http://duckduckgo.com",
    )
    response = {
        "backend": backend,
        "body": "Hey, this isn't http://google.com!",
    }
    httpserver.expect_request("/{}".format(backend)).respond_with_data(
        salt.utils.json.dumps(response), content_type="text/plain"
    )
    rendered = render_jinja_tmpl(
        "{{ '"
        + httpserver.url_for("/{}".format(backend))
        + "' | http_query(backend='"
        + backend
        + "') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert isinstance(rendered, str), "Failed with rendered template: {}".format(
        rendered
    )
    dict_reply = ast.literal_eval(rendered)
    assert isinstance(dict_reply, dict), "Failed with rendered template: {}".format(
        rendered
    )
    assert (
        "body" in dict_reply
    ), "'body' not found in request response({}). Rendered template: {!r}".format(
        dict_reply, rendered
    )
    assert isinstance(
        dict_reply["body"], str
    ), "Failed with rendered template: {}".format(rendered)


def test_to_bool(minion_opts, local_salt):
    """
    Test the `to_bool` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 1 | to_bool }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "True"

    rendered = render_jinja_tmpl(
        "{{ 'True' | to_bool }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "True"

    rendered = render_jinja_tmpl(
        "{{ 0 | to_bool }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "False"

    rendered = render_jinja_tmpl(
        "{{ 'Yes' | to_bool }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "True"


def test_quote(minion_opts, local_salt):
    """
    Test the `quote` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'random' | quote }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "random"


def test_regex_search(minion_opts, local_salt):
    """
    Test the `regex_search` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'abcdefabcdef' | regex_search('BC(.*)', ignorecase=True) }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "('defabcdef',)"  # because search looks only at the beginning


def test_regex_match(minion_opts, local_salt):
    """
    Test the `regex_match` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'abcdefabcdef' | regex_match('BC(.*)', ignorecase=True)}}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "None"


def test_regex_replace(minion_opts, local_salt):
    """
    Test the `regex_replace` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        r"{{ 'lets replace spaces' | regex_replace('\s+', '__') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "lets__replace__spaces"


def test_uuid(minion_opts, local_salt):
    """
    Test the `uuid` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'random' | uuid }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "3652b285-26ad-588e-a5dc-c2ee65edc804"


def test_min(minion_opts, local_salt):
    """
    Test the `min` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, 3] | min }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "1"


def test_max(minion_opts, local_salt):
    """
    Test the `max` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, 3] | max }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "3"


def test_avg(minion_opts, local_salt):
    """
    Test the `avg` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, 3] | avg }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "2.0"


def test_union(minion_opts, local_salt):
    """
    Test the `union` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, 3] | union([2, 3, 4]) | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "1, 2, 3, 4"


def test_intersect(minion_opts, local_salt):
    """
    Test the `intersect` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, 3] | intersect([2, 3, 4]) | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "2, 3"


def test_difference(minion_opts, local_salt):
    """
    Test the `difference` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, 3] | difference([2, 3, 4]) | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "1"


def test_symmetric_difference(minion_opts, local_salt):
    """
    Test the `symmetric_difference` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, 3] | symmetric_difference([2, 3, 4]) | join(', ') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "1, 4"


def test_method_call(minion_opts, local_salt):
    """
    Test the `method_call` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 6|method_call('bit_length') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "3"
    rendered = render_jinja_tmpl(
        "{{ 6.7|method_call('is_integer') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "False"
    rendered = render_jinja_tmpl(
        "{{ 'absaltba'|method_call('strip', 'ab') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "salt"
    rendered = render_jinja_tmpl(
        "{{ [1, 2, 1, 3, 4]|method_call('index', 1, 1, 3) }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "2"

    # have to use `dictsort` to keep test result deterministic
    rendered = render_jinja_tmpl(
        "{{ {}|method_call('fromkeys', ['a', 'b', 'c'], 0)|dictsort }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "[('a', 0), ('b', 0), ('c', 0)]"

    # missing object method test
    rendered = render_jinja_tmpl(
        "{{ 6|method_call('bit_width') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "None"


def test_md5(minion_opts, local_salt):
    """
    Test the `md5` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'random' | md5 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "7ddf32e17a6ac5ce04a8ecbf782ca509"


def test_sha256(minion_opts, local_salt):
    """
    Test the `sha256` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'random' | sha256 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert (
        rendered == "a441b15fe9a3cf56661190a0b93b9dec7d04127288cc87250967cf3b52894d11"
    )


def test_sha512(minion_opts, local_salt):
    """
    Test the `sha512` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'random' | sha512 }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == str(
        "811a90e1c8e86c7b4c0eef5b2c0bf0ec1b19c4b1b5a242e6455be93787cb473cb7bc"
        "9b0fdeb960d00d5c6881c2094dd63c5c900ce9057255e2a4e271fc25fef1"
    )


def test_hmac(minion_opts, local_salt):
    """
    Test the `hmac` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'random' | hmac('secret', 'blah') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "False"

    rendered = render_jinja_tmpl(
        "{{ 'get salted' | "
        "hmac('shared secret', 'eBWf9bstXg+NiP5AOwppB5HMvZiYMPzEM9W5YMm/AmQ=') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "True"


def test_base64_encode(minion_opts, local_salt):
    """
    Test the `base64_encode` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'random' | base64_encode }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "cmFuZG9t"


def test_base64_decode(minion_opts, local_salt):
    """
    Test the `base64_decode` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ 'cmFuZG9t' | base64_decode }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "random"


def test_json_query(minion_opts, local_salt):
    """
    Test the `json_query` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, 3] | json_query('[1]')}}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "2"


def test_flatten_simple(minion_opts, local_salt):
    """
    Test the `flatten` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, [3]] | flatten }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "[1, 2, 3]"


def test_flatten_single_level(minion_opts, local_salt):
    """
    Test the `flatten` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, [None, 3, [4]]] | flatten(levels=1) }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "[1, 2, 3, [4]]"


def test_flatten_preserve_nulls(minion_opts, local_salt):
    """
    Test the `flatten` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ [1, 2, [None, 3, [4]]] | flatten(preserve_nulls=True) }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "[1, 2, None, 3, 4]"


def test_dict_to_sls_yaml_params(minion_opts, local_salt):
    """
    Test the `dict_to_sls_yaml_params` Jinja filter.
    """
    expected = [
        "- name: donkey",
        "- list:\n  - one\n  - two",
        "- dict:\n    one: two",
        "- nested:\n  - one\n  - two: three",
    ]
    source = (
        "{% set myparams = {'name': 'donkey', 'list': ['one', 'two'], 'dict': {'one': 'two'}, 'nested': ['one', {'two': 'three'}]} %}"
        + "{{ myparams | dict_to_sls_yaml_params }}"
    )
    rendered = render_jinja_tmpl(
        source, dict(opts=minion_opts, saltenv="test", salt=local_salt)
    )
    assert rendered in ["\n".join(combo) for combo in itertools.permutations(expected)]


def test_combinations(minion_opts, local_salt):
    """
    Test the `combinations` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{% for one, two in 'ABCD' | combinations(2) %}{{ one~two }} {% endfor %}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "AB AC AD BC BD CD "


def test_combinations_with_replacement(minion_opts, local_salt):
    """
    Test the `combinations_with_replacement` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{% for one, two in 'ABC' | combinations_with_replacement(2) %}{{ one~two }} {% endfor %}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "AA AB AC BB BC CC "


def test_compress(minion_opts, local_salt):
    """
    Test the `compress` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{% for val in 'ABCDEF' | compress([1,0,1,0,1,1]) %}{{ val }} {% endfor %}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "A C E F "


def test_permutations(minion_opts, local_salt):
    """
    Test the `permutations` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{% for one, two in 'ABCD' | permutations(2) %}{{ one~two }} {% endfor %}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "AB AC AD BA BC BD CA CB CD DA DB DC "


def test_product(minion_opts, local_salt):
    """
    Test the `product` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{% for one, two in 'ABCD' | product('xy') %}{{ one~two }} {% endfor %}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "Ax Ay Bx By Cx Cy Dx Dy "


def test_zip(minion_opts, local_salt):
    """
    Test the `zip` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{% for one, two in 'ABCD' | zip('xy') %}{{ one~two }} {% endfor %}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "Ax By "


def test_zip_longest(minion_opts, local_salt):
    """
    Test the `zip_longest` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{% for one, two in 'ABCD' | zip_longest('xy', fillvalue='-') %}{{ one~two }} {% endfor %}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "Ax By C- D- "


def test_random_sample(minion_opts, local_salt):
    """
    Test the `random_sample` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ ['one', 'two', 'three', 'four'] | random_sample(2, seed='static') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "['four', 'two']"


def test_random_shuffle(minion_opts, local_salt):
    """
    Test the `random_shuffle` Jinja filter.
    """
    rendered = render_jinja_tmpl(
        "{{ ['one', 'two', 'three', 'four'] | random_shuffle(seed='static') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == "['four', 'two', 'three', 'one']"


def test_ifelse(minion_opts, local_salt):
    """
    Test the `ifelse` Jinja global function.
    """
    rendered = render_jinja_tmpl(
        "{{ ifelse('default') }}\n"
        "{{ ifelse('foo*', 'fooval', 'bar*', 'barval', 'default', minion_id='foo03') }}\n"
        "{{ ifelse('foo*', 'fooval', 'bar*', 'barval', 'default', minion_id='bar03') }}\n"
        "{{ ifelse(False, 'fooval', True, 'barval', 'default', minion_id='foo03') }}\n"
        "{{ ifelse('foo*', 'fooval', 'bar*', 'barval', 'default', minion_id='baz03') }}",
        dict(opts=minion_opts, saltenv="test", salt=local_salt),
    )
    assert rendered == ("default\n" "fooval\n" "barval\n" "barval\n" "default")



from functools import partial

import pytest

import salt.ext.tornado.gen
from salt.netapi.rest_tornado import saltnado

# TODO: run all the same tests from the root handler, but for now since they are
# the same code, we'll just sanity check


@pytest.fixture
def app_urls():
    return [
        (r"/events", saltnado.EventsSaltAPIHandler),
    ]


@pytest.mark.slow_test
async def test_get(http_client, io_loop, app):
    events_fired = []

    def on_event(events_fired, event):
        if len(events_fired) < 6:
            event = event.decode("utf-8")
            app.event_listener.event.fire_event(
                {"foo": "bar", "baz": "qux"}, "salt/netapi/test"
            )
            events_fired.append(1)
            event = event.strip()
            # if we got a retry, just continue
            if event != "retry: 400":
                tag, data = event.splitlines()
                assert tag.startswith("tag: ")
                assert data.startswith("data: ")

    # We spawn the call here because otherwise the fetch method would
    # continue reading indefinitely and there would be no wait to
    # properly run the assertions or stop the request.
    io_loop.spawn_callback(
        http_client.fetch,
        "/events",
        streaming_callback=partial(on_event, events_fired),
        request_timeout=30,
    )

    while len(events_fired) < 5:
        await salt.ext.tornado.gen.sleep(1)

    assert len(events_fired) >= 5
